/*
  file:         CreateMatrix.cc
  created:      David P Jacobs May 1990
  modified:     pasha jun 18 2019
  modification: some funcs turned to Albert methods
*/

// This module contains routines dealing with creating a Matrix for the given 
// list of equations (i.e. list Basis pairs)/
//    1) Initialize PairPresent bit Matrix to 0's.
//    2) Compute 'N', the number of distinct Basis pairs present in the given 
//       list of equations.
//    3) malloc ColtoBP array of 'N' pairs w.o coefficients.
//    4) Fill the ColtoBP array with the distinct Bais pairs.
//       Use the PairPresent bit Matrix while filling.
//    5) Allocate space for the Matrix of equations.
//       It's number of rows is equal to number of equations.
//       And number of columns equal to 'N'.
//       ZeroOut the Matrix.
//       For each equation, one row of the Matrix is filled.
//       For each Basis pair in the equation, scan the array to find the column
//       number. It is nothing but the position where we could find the required
//       Basis pair.
//       Enter the Scalar corresponding to the Basis pair in the equaition into
//       the Matrix in the corresponding column.

#include <iostream>
#include "albert.hh"

#include "CreateMatrix.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

bool AreBasisElements (const Degree d);



int Albert::DestroyBPtoCol()
{
  assert_not_null(this->ColtoBP_CreateMatrix);
  free (this->ColtoBP_CreateMatrix);
  return (1);
}


bool Albert::CreateColtoBP()
{
  if (this->Pair_present.size() == 0)
	{
	  return (1);
	}
  this->ColtoBP_CreateMatrix = 
	(Unique_basis_pair_list)(Mymalloc ((this->Pair_present.size()) *
									   sizeof(Unique_basis_pair)));
  assert_not_null(this->ColtoBP_CreateMatrix);
  const Degree d = GetDegreeName (this->N_CreateMatrix);
  int col_to_bp_index = 0;
  for (int i=1; i<d; i++)
	{
	  if (AreBasisElements (i) && AreBasisElements (d-i)) 
		{
		  this->Process (i, d-i, &col_to_bp_index);
		}
	}
  return (1);
}



// Returns true if there are basis elements at degree d.
// Note: when the algebra is nilpotent it is possible that no new basis 
// elements were entered at degree d.
bool Albert::AreBasisElements (const Degree d) const
{
  return (this->BasisStart (d) != 0);
}



void Albert::Process 
(const Degree d1, const Degree d2, int* col_to_bp_index_ptr)
{
  const Basis b1 = this->BasisStart (d1);
  const Basis b2 = this->BasisEnd   (d1);
  const Basis b3 = this->BasisStart (d2);
  const Basis b4 = this->BasisEnd   (d2);
  Basis j;
  for (Basis i = b1; i <= b2; i++) 
	{
	  for (j = b3; j <= b4; j++) 
		{
		  if (this->Pair_present.find (std::make_pair (i, j)) != 
			  this->Pair_present.end())
			{
			  this->ColtoBP_CreateMatrix[*col_to_bp_index_ptr].left_basis  = i;
			  this->ColtoBP_CreateMatrix[*col_to_bp_index_ptr].right_basis = j;
			  (*col_to_bp_index_ptr)++;
            }
        }
    }
} // end of Albert::Process()

            



int Albert::GetCol (const Basis Left_basis, const Basis Right_basis) const
{
  if ((this->Pair_present.size() == 0) || (this->Num_equations == 0))
	{
	  return (-1);
	}
  int low = 0;
  int high = this->Pair_present.size() - 1;
  int middle;
  while (low <= high) 
	{
	  middle = (low + high)/2; 
	  if (Left_basis < this->ColtoBP_CreateMatrix[middle].left_basis)
		{
		  high = middle - 1;
		}
	  else if (Left_basis > this->ColtoBP_CreateMatrix[middle].left_basis)
		{
		  low = middle + 1;
		}
	  else if (Right_basis < this->ColtoBP_CreateMatrix[middle].right_basis)
		{
		  high = middle - 1;
		}
	  else if (Right_basis > this->ColtoBP_CreateMatrix[middle].right_basis)
		{
		  low = middle + 1;
		}
	  else
		{
		  return (middle);
		}
    }
  return (-1);
} // end of Albert::GetCol()



// end of CreateMatrix.cc
