/*
  file:         GFp.hh (former Scalar_arithmetic.hh)
  purpose:      implements GF(p), p>2
  created:      Pasha Zusmanovich mar 8 2009
  modified:     pasha jun 16,19,26,28-29 2019
  modification: - removed DEFAUL_CHAR
                - some operators transferred to base class
				- operator<< transferred to .cc file
*/

#ifndef __GFP_H
#define __GFP_H

#include <stdlib.h>
//#include <iostream>
#include <map>
#include "scalar.hh"
#include "Debug.hh"

char      _abs    (long long&);
long long _reduce (long long);
long long _S_add  (const long long, const long long);
long long _S_mul2 (const long long&);

//#define DEFAULT_CHAR 251


class GFp_element: public Scalar<long long>
{
public:
  static unsigned long long HALF_PRIME;

  // to keep inverse of elements in the range [1, (PRIME-1)/2]
  static std::map<unsigned long long, long long> Inverse_table;

  inline GFp_element () {}

  inline GFp_element (const long long i)
  {
	// TODO: more efficient?
	this->val = _reduce (i % (long long)Albert::PRIME);
  }


  friend inline GFp_element operator- (const GFp_element& X)
  {
#if DEBUG_MODULAR
	_scalar_assert (x);
#endif
	GFp_element M;
	M.val = -X.val;
	return (M);
  }


  friend inline GFp_element operator+ (const GFp_element& X, 
									   const GFp_element& Y)
  {
	GFp_element S;
	S.val = _S_add (X.val, Y.val);
	return (S);
  }
  


  friend inline GFp_element operator* (const GFp_element& X, 
									   const GFp_element& Y)
  {
	/*
	  x = 101101;
	  x*y = x*10000 + x*1000 + ...
	  x*2^n + x*2^k + ...
	*/
#if DEBUG_MODULAR
	_scalar_assert (X);
	_scalar_assert (Y);
#endif
	
	long long x = X.val;
	long long y = Y.val;
	char sign = _abs(x) * _abs(y);
	if (x < y)
	  {
		long long tmp = x;
		x = y;
		y = tmp;
	  }
	
	// get the rightmost binary digit of y:
	long long x2n = x;    // x * 2^n
	long long result = (y & 0x1 ? x2n : 0);
	
	// loop thru the rest of digits of y, from the right
	while (1)
	  {
		y = y >> 1;
		if (y == 0)
		  {
			break;
		  }
		x2n = _S_mul2 (x2n);
		if (y & 0x1)
		  {
			result = _S_add (result, x2n);
		  }
	  }
#if DEBUG_MODULAR  
	//std::cerr << X << " * " << Y << " = " << (sign == 1 ? result : -result) << std::endl;
#endif
	GFp_element P;
	P.val = (sign == 1 ? result : -result);
	return (P);
  } // end of operator*




  friend inline GFp_element S_inv (const GFp_element& X)
  {
#if DEBUG_MODULAR
	_scalar_assert (X);
#endif
	long long x = X.val;
	if (x == 0)
	  { 
		std::cerr << "ERROR: division by 0 in S_inv()" << std::endl;
		exit (-1);
	  }
	const char sign = _abs (x);
	
	// check whether this value was inverted already;
	// we search only for values in the range [1, (PRIME-1)/2]
	long long D;
	GFp_element::ITI = GFp_element::Inverse_table.find (x);
	if (GFp_element::ITI == GFp_element::Inverse_table.end())
	  {
		// compute 1/x (mod PRIME);
		// modified from LibTomMath (http://math.libtomcrypt.org/), 
		// file bn_fast_mp_invmod.c; 
		// this is basically a binary extended gcd algorithm from 
		// Knuth, Vol. 2, 3rd ed., Exercise 4.5.2.39 (p. 646), 
		// with minor modifications to prevent integers overflow;
		// this is *not* good for PRIME == 2
		long long u = Albert::PRIME;
		long long v = x;
		long long B = 0;
		D = 1;
		
	  top:
		while (! (u & 0x1))  // while u is even
		  {
			u = u >> 1;  // u = u/2
			if (B & 0x1) // B is odd
			  {
				// integer overflow should never occur here,
				// as at this point |B| < PRIME
				if (B > 0)
				  {
					B -= Albert::PRIME;
				  }
				else
				  {
					B += Albert::PRIME;
				  }
			  }
			B = B >> 1;  // B = B/2
			// at this point, |B| < PRIME/2
		  }
		while (! (v & 0x1))  // while v is even
		  {
			v = v >> 1;  // v = v/2
			if (D & 0x1) // D is odd
			  {
				// ditto about integer overflow
				if (D > 0)
				  {
					D -= Albert::PRIME;
				  }
				else
				  {
					D += Albert::PRIME;
				  }
			  }
			D = D >> 1;  // D = D/2
			// ditto for |D|
		  }
		
		if (u >= v)
		  {
			// integer overflow should never occur here,
			// as initially u = PRIME, and u is non-negative and
			// non-increasing on each step
			u -= v;
			// at this point, |B|,|D| < PRIME/2, hence ...
			B -= D;
			// ... at this point |B| < PRIME
		  } 
		else 
		  {
			// integer overflow should never occur here,
			// as initially v < PRIME/2, and v is non-negative and
			// non-increasing on each step
			v -= u;
			D -= B;
			// ditto for |D|
		  }
		
		//std::cout << "B=" << B << " D=" << D << " u=" << u << " v=" << v << std::endl;
		
		if (u != 0)
		  {
			goto top;
		  }
		
		// at this point it should be:
		// v == 1 (violated if PRIME is not prime)
		// and 
		// B == 0 (mod PRIME) (violated if there is integer overflow while 
		// performing subtractions above)
#if DEBUG_MODULAR
		if (v != 1)
		  {
			std::cerr << Albert::PRIME << " is not a prime in S_inv(): v=" 
					  << v << std::endl;
			exit (-1);
		  }
		if ((B % (long long)Albert::PRIME) != 0)
		  {
			std::cerr
			  << "integer overflow in S_inv(): PRIME=" << Albert::PRIME 
			  << " B=" << B << std::endl;
			exit (-1);
		  }
#endif
		
		if (D < -(long long)GFp_element::HALF_PRIME)
		  {
			D += Albert::PRIME;
		  }
		else if (D > (long long)GFp_element::HALF_PRIME)
		  {
			D -= Albert::PRIME;
		  }
		GFp_element::Inverse_table.insert (std::make_pair (x, D));
		// note that we insert positive keys only
		if (D > 0)
		  {
			GFp_element::Inverse_table.insert (std::make_pair (D, x));
		  }
		else
		  {
			GFp_element::Inverse_table.insert (std::make_pair (-D, -x));
		  }
#if DEBUG_MODULAR
		//std::cerr << "1/" << X << " = " << (sign == 1 ? D : -D) << std::endl;
#endif
	  }
	else
	  {
#if DEBUG_MODULAR
		//std::cerr << "using inverse table for X=" << X << std::endl;
#endif
		D = GFp_element::ITI->second;
	  }

	GFp_element I;
	I.val = (sign == 1 ? D : -D);
	return (I);
  } // end of S_inv()


  inline GFp_element& operator*= (const GFp_element& X)
  {
	*this = *this * X;
	return (*this);
  }


  // overrides the base class, defined in the .cc file
  friend std::ostream& operator<< (std::ostream&, const GFp_element&);


private:
  // the value is stored in the range [-(PRIME-1)/2, (PRIME-1)/2];

  // to search in the Inverse_table
  static std::map<unsigned long long, long long>::const_iterator ITI;

}; // end of class GFp_element





// i assumed to be in the range [-PRIME, PRIME];
// reduce to the range [-(PRIME-1)/2, (PRIME-1)/2]
inline void _reduce_in_place (long long& i)
{
  if (i < -(long long)GFp_element::HALF_PRIME)
    {
      i += Albert::PRIME;
    }
  else if (i > (long long)GFp_element::HALF_PRIME)
    {
      i -= Albert::PRIME;
    }
}


inline long long _reduce (long long i)
{
  _reduce_in_place (i);
  return (i);
}



inline long long _S_add (const long long x, const long long y)
{
#if DEBUG_MODULAR
  _scalar_assert (x);
  _scalar_assert (y);
  //std::cerr << x << " + " << y << " = " << _reduce (x+y) << std::endl;
#endif
  return (_reduce (x+y));
}




// absolute value: returns: sign (i.e. 1 the initial value is positive, -1 if negative)
inline char _abs (long long& x)
{
  if (x < 0)
    {
      x = -x;
      return (-1);
    }
  else
    {
      return (1);
    }
}


// multiplication by 2
inline long long _S_mul2 (const long long& x)
{
#if DEBUG_MODULAR
  //std::cerr << x << " times 2 = " << _reduce(x << 1) << std::endl;
#endif
  return (_reduce (x << 1));
}


#endif // __GFP_H
// end of GFp.hh
