/*
  file:         GenerateEquations.cc
  created:      David P Jacobs circa 1990
  modified:     pasha jun 19 2019
  modification: Cur_deg_vars turned to std::vector
*/

// Given an identity and type, this module generates all the equations by making
// calls to other modules.
// For each type generate Sequential subtypes, i.e. given a type with l 
// generators, an identity F with m variables, say x1,...,xm, variable xi having
// degree di, we generate sequential subtypes, t1,...,tm such that
// t1 + ... + tm = t   and   deg(ti) >= deg(xi).
//
//        Type t    1  2  . . . . . . . . . . . l                
//                  -----------------------------             
//                 |  |  |                    |  |            
//                  -----------------------------             
//
//                  -----------------------------             
//             t1  |  |  |                    |  |  >=d1      
//                  -----------------------------             
//             t2  |  |  |                    |  |  >=d2      
//                  -----------------------------             
//                 |  |  |                    |  |            
//                  -----------------------------             
//                 |  |  |                    |  |            
//                  -----------------------------             
//                 |  |  |                    |  |            
//                  -----------------------------             
//             tm  |  |  |                    |  |  >=dm      
//                  -----------------------------             
//
// e.g. F(x,y) = (xx)(xy) - ((xy)x)x.                       
// So degree of x = 3 and degree of y = 1.                
// Let the target type be (3,2,1).                        
// Suppose we are constructing degree 5 basis words.      
// Then degree 5 subtypes are (2,2,1), (3,1,1) and (3,2,0).
//
// Let t = (2,2,1)                                         
//                               1  2   3                         
//                              -----------                       
//                             | 2 | 2 | 1 |                      
//                              -----------                       
//                                                              
// Sequential subtypes are :                            
//
//            -----------      -----------      -----------       
//        x  | 1 | 2 | 1 |  x | 2 | 1 | 1 |  x | 2 | 2 | 0 |      
//            -----------      -----------      -----------       
//        y  | 1 | 0 | 0 |  y | 0 | 1 | 0 |  y | 0 | 0 | 1 |      
//            -----------      -----------      -----------       
//                                                              
//            -----------      -----------      -----------       
//        x  | 0 | 2 | 1 |  x | 2 | 0 | 1 |  x | 1 | 1 | 1 |      
//            -----------      -----------      -----------       
//        y  | 2 | 0 | 0 |  y | 0 | 2 | 0 |  y | 1 | 1 | 0 |      
//            -----------      -----------      -----------       


#include <iostream>
#include "albert.hh"

#include "GenerateEquations.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif


void Albert::InitSeqSubtypes()
{
  this->Whatsleft = this->Target_type_deg;
  int i, j;
  for (i=0; i < this->Num_vars; i++)
	{
	  //this->Cur_deg_vars[i] = 0; 
	  this->Cur_deg_vars.push_back (0); 
	}
  for (i=0; i < this->Num_vars; i++)
	{
	  for (j=0; j < this->Target_type_len; j++)
		{
		  this->Seq_sub_types[i * (this->Target_type_len) + j] = 0;
		}
	}
}





int Albert::GetVarNumber (const char Letter) const
{
  int var_num = 1;
  for (int i=0; i<NUM_LETTERS; i++) 
	{
	  if ((Letter - 'a') == i)
		{
		  return (var_num);
		}
	  else if (this->The_ident->deg_letter[i] > 0)
		{
		  var_num++;
		}
    }
  std::cout << "ERROR from GetVarNumber()" << std::endl;
  return (-1);
  // or should it be just return (var_num)?
} 



#if DEBUG_SEQ_SUBTYPES
void Albert::PrintSeqSubtypes() const
{
  int j;
  for (int i=0; i < this->Num_vars; i++) 
	{
	  std::cout << "    ";
	  for (j=0; j < this->Target_type_len; j++)
		{
		  //printf("%d",Seq_sub_types[i*Target_type_len) + j]);
		  std::cout << this->Seq_sub_types[i * (this->Target_type_len) + j];
		}
	  std::cout << std::endl;
	}
}
#endif




// end of GenerateEquations.cc
