/*
  file:         Multpart.cc
  created:      David P Jacobs circa 1990
  modified:     pasha jun 19,aug 4 2019
  modification: eliminated Deg_var_types and Deg_vars
*/


/***      The idea of set partitioning is to linearize the         ***/ 
/***      identity.                                                ***/ 
/***      Each sequential subtype tj is partitioned into dj parts  ***/
/***      such that     tj = tj1 + tj2 + .... + tjdj    where      ***/
/***      dj is the degree of a jth variable.                      ***/
/***                                                               ***/
/***                 t1     t2                 tm                  ***/ 
/***                ------------------------------                 ***/
/***   Sequential  |     |     |            |     |                ***/ 
/***   subtype tj   ------------------------------                 ***/
/***                                                               ***/
/***                                                               ***/
/***                ------------------------------                 ***/
/***          t11  |     |     |            |     |                ***/ 
/***                ------------------------------                 ***/
/***          t12  |     |     |            |     |                ***/ 
/***                ------------------------------                 ***/
/***               |     |     |            |                      ***/ 
/***                ------------------------                       ***/
/***               |     |     |            |                      ***/
/***                ------------------------                       ***/
/***          t1d1 |     |     |            |                      ***/ 
/***                -----       ------------                       ***/
/***                                                               ***/
/***                                                               ***/
/***      e.g F(x,y) = (xx)(xy) - ((xy)x)x.                        ***/
/***      So degree of x = 3 and degree of y = 1.                  ***/
/***      Let the sequential subtype be                            ***/ 
/***                                                               ***/
/***                   x        y                                  ***/
/***                ---------------                                ***/
/***               | 1 2 1 | 1 0 0 |                               ***/
/***                ---------------                                ***/
/***                                                               ***/
/***                                                               ***/
/***         Set partitions are :                                  ***/
/***                                                               ***/
/***     ---------------     ---------------     ---------------   ***/ 
/***  u | 1 0 0 | 1 0 0 |   | 1 1 0 | 1 0 0 |   | 0 1 1 | 1 0 0 |  ***/
/***     ---------------     ---------------     ---------------   ***/ 
/***  v | 0 2 0 |           | 0 1 0 |           | 1 0 0 |          ***/
/***     -------             -------             -------           ***/ 
/***  w | 0 0 1 |           | 0 0 1 |           | 0 1 0 |          ***/
/***     -------             -------             -------           ***/ 
/***                                                               ***/


#include "albert.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif



void Albert::AddSetPartition (const Name n, const int j)
{
  this->Set_partitions[this->Cur_index_var[j] * (this->Num_vars) + j] = n;
  this->Cur_index_var[j]++;
}


void Albert::DeleteSetPartition (const Name j)
{
  --(this->Cur_index_var[j]);
}


bool Albert::OKSetPartitions() const
{
  int i;
  for (int j=0; j < this->Num_vars; j++)
	{
	  for (i=0; i < (this->Deg_var[j] - 1); i++)
		{
		  if (this->Set_partitions[    i * (this->Num_vars) + j] < 
			  this->Set_partitions[(i+1) * (this->Num_vars) + j])
			{
			  return (0);
			}
		}
	}
  return (1);
}


#if DEBUG_SET_PARTITIONS
void Albert::PrintVarTypes() const
{
  std::cout << "Variable Types are : " << std::endl;
  for (int i=0; i < this->Num_vars; i++) 
	{
	  PrintTypeName (this->Var_types[i], std::cout);
	  std::cout << std::endl;
    }
}

void Albert::PrintSetPartitions() const
{
  static int count = 0;
  std::cout << count++ << "th partition is " << std::endl;
  int i;
  for (int j=0; j < this->Num_vars; j++)
	{
	  for (i=0; i < this->Deg_var[j]; i++) 
		{
		  PrintTypeName 
			(this->Set_partitions[i * (this->Num_vars) + j], std::cout);
		  std::cout << std::endl;
		}
	}
}
#endif


// end of Multpart.cc
