/*
  file:         Po_routines.cc
  purpose:      routines dealing with printing the polynomial and 
                finding if a polynomial is homogeneous or not
  created:      Sekhar Muddana circa 1990
  modified:     pasha jun 19 2019
  modification: minor cleanup
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include "albert.hh"
#include "assert_not_null.hh"

#include "Po_routines.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

int  Absolute (int);
void AssignNumbersToTerm (term_node*, int[]);
void Create_term_str     (term_node*, char[]);
void DestroyTerms        (term_head*);
void FreeNodes           (term_node*);
unsigned int Get_len     (const int);
int  Get_max_coef        (term_head*);
bool Same_type           (const P_type&, const P_type&);
void Store_type          (term_node*, P_type*);



// Print the polynomial in nice looking format, where all terms are properly 
// aligned.                                
// This routine is called when a command "d" is issued.
// arg: Poly -- Pointer to a polynomial to be printed
void Print_poly (const polynomial* Poly, const size_t Poly_len)
{
  if (Poly == NULL)
	{
	  std::cout << "Print_poly(Poly): Poly is null pointer" << std::endl;
	}
  else 
	{
	  term_head* temp_head = Poly->terms;
	  int max_coef = 0;
	  unsigned int len_max_coef = 0;
	  if (temp_head != NULL) 
		{
		  max_coef = Get_max_coef (temp_head);
		  len_max_coef = Get_len (max_coef);
		}

	  char* Term_str = (char*)Mymalloc (Poly_len+1);
	  Term_str[0] = '\0';
	  
	  unsigned int term_len;
	  bool first_term = 1;
	  unsigned int cur_col = 0; 
	  unsigned int cur_row = 0;
	  unsigned int len_coef = 0;
	  unsigned int i;
   	  while (temp_head != NULL) 
		{
		  Create_term_str (temp_head->term, Term_str);
		  term_len = strlen (Term_str);

		  cur_col += 3 + len_max_coef + term_len;

		  // Get rid of extra '(' in the beginning and extra ')' in the end
		  // of each term.
		  if (term_len > 2) 
			{
			  Term_str[0] = Term_str[term_len - 1] = ' ';
            }

		  if (cur_col > 80) 
			{
			  ++cur_row;
			  std::cout << std::endl;
			  cur_col = 3 + len_max_coef + term_len;
            }

		  if((temp_head->coef >= 1) && (! first_term))
			{
			  std::cout << " + ";
			}
		  else if (temp_head->coef <= -1)
			{
			  std::cout << " - ";
			}
		  if (first_term) 
			{
			  first_term = 0;
			  if (temp_head->coef > 0)
				{
				  std::cout << "   ";
				}
            }

		  len_coef = Get_len (Absolute (temp_head->coef));
		  for (i = len_coef; i < len_max_coef; i++)
			{
			  std::cout << " ";
			}
		  if (temp_head->coef > 1)
			{
			  std::cout << temp_head->coef;
			}
		  else if (temp_head->coef < -1)
			{
			  std::cout << -temp_head->coef;
			}
            else 
			  {
				for (i = 0; i < len_max_coef; i++)
				  {
					std::cout << " ";
				  }
			  }

		  if (term_len > 2)
			{
			  std::cout << Term_str;
			}
		  else
			{
			  std::cout << " " << Term_str << " ";
			}
		  Term_str[0] = '\0';

		  temp_head = temp_head->next;
        }
	  free (Term_str);
    }
} // end of Print_poly()





// RETURNS: Absolute value of Num
int Absolute (int Num)
{
  /*
  if (Num < 0)
	{
	  return (-Num);
	}
  else
	{
	  return (Num);
	}
  */
  return ((Num < 0) ? -Num : Num);
}



// RETURNS: Number of digits in the integer Num
unsigned int Get_len (const int Num)
{
  unsigned int i = 1;
  int temp = Num;
  while ((temp /= 10) > 0)
	{
	  i++;
	}
  return (i);
}



// args: Pntr -- Pointer to the Head of list of terms in polynomial
// RETURNS: Number of digits in the largest coefficient of all terms in 
//          a  polynomial.
int Get_max_coef (term_head* Pntr)
{
  if (Pntr == NULL)
	{
	  return (0);
	}
  term_head* temp = Pntr;
  int max = 0;
  while (temp != NULL) 
	{
	  if (Absolute(temp->coef) > max)
		{
		  max = Absolute (temp->coef);
		}
	  temp = temp->next;
    }
  return (max);
}



// Traverse the tree pointed to by Pntr recursively, attaching characters 
// (i.e small letters or '(' or ')') to the Term_str.
// MODIFIES: Term_str -- string which contains string equivalent of a term_node
//           passed when control returns to the caller.
// arg: Pntr -- Pointer to a term_node of a polynomial.
void Create_term_str (term_node* Pntr, char Term_str[])
{
  char temp_str[2];

  if (((Pntr->left) == NULL) && ((Pntr->right) == NULL)) 
	{
	  sprintf (temp_str, "%c",Pntr->letter);
	  strcat (Term_str, temp_str);
#if DEBUG_ASSIGN_NUMBERS
	  sprintf (temp_str, "%d", Pntr->number);
	  strcat (Term_str, temp_str);
#endif
	}
  else 
	{
	  sprintf (temp_str, "(");
	  strcat (Term_str, temp_str);
	  Create_term_str (Pntr->left, Term_str);
	  Create_term_str (Pntr->right, Term_str);
	  sprintf (temp_str, ")");
	  strcat (Term_str, temp_str);
    }
}




// Check if the Polynomial is homogeneous by constructing the type of first term
// and comparing it's type with all other terms in the polynomial.
// All terms in a Polynomial will have the same type if it is homogeneous. i.e 
// All terms will have same small letters and degree of each small letter is 
// same in all terms of the Polynomial.
// If the Polynomial is homogeneous, its type will be stored in the Polynomial.
// arg: Pntr -- Pointer to a polynomial.              
// RETURNS: 1 if the Polynomial is homogeneous, 0 otherwise.
bool Homogeneous (polynomial* Poly)
{
  const term_head* temp_head;
  P_type first_term_type, temp_type;
  bool is_homogeneous = 1;
  unsigned int i;
  
  for (i=0; i<NUM_LETTERS; i++) 
	{
	  first_term_type.degrees[i] = 0;
	  temp_type.degrees[i] = 0;
    }
  first_term_type.tot_degree = 0;
  temp_type.tot_degree = 0;
  
  temp_head = Poly->terms;
  if (temp_head != NULL)
	{
	  Store_type (temp_head->term, &first_term_type);
	}
  
  temp_head = temp_head->next;
  while (temp_head != NULL) 
	{
	  Store_type (temp_head->term, &temp_type);
	  if (! Same_type (first_term_type, temp_type)) 
		{
		  is_homogeneous = 0;
		  break;
		}
	  for (i=0; i<NUM_LETTERS; i++)
		{
		  temp_type.degrees[i] = 0;
		}
	  temp_type.tot_degree = 0;
	  temp_head = temp_head->next;
    }
  
  if (! is_homogeneous)
	{
	  return (0);
	}
  else 
	{
	  for (i=0; i<NUM_LETTERS; i++)
		{
		  Poly->deg_letter[i] = first_term_type.degrees[i];
		}
	  Poly->degree = first_term_type.tot_degree;
	  return (1);
    }
} // end of Homogeneous()





// Increment the degree of small letters traversing the tree pointed to by term
// recursively.                           
// MODIFIES: type -- the type of the polynomial is stored. 
// args: term -- Pointer to the term tree (of term_node's) whose type is to be 
//               computed.
void Store_type (term_node* term, P_type* type)
{
  if ((term->left == NULL) && (term->right == NULL)) 
	{
	  type->degrees[term->letter - 'a'] += 1;
	  type->tot_degree += 1;
    }
  else 
	{
	  Store_type (term->left,  type);
	  Store_type (term->right, type);
    }
}




// RETURNS: 1 if type1 and type2 have same degree in each small letter,
//          0 otherwise.
bool Same_type (const P_type& type1, const P_type& type2)
{
  bool is_sametype = 1;
  for (unsigned int i=0; i<NUM_LETTERS; i++) 
	{
	  if (type1.degrees[i] != type2.degrees[i]) 
		{
		  is_sametype = 0;
		  break;
		}
    }
  /*
  if (! is_sametype)
	{
	  return (0);
	}
  else
	{
	  return (1);
	}
  */
  return (is_sametype);
}




// arg: Poly -- Pointer to a polynomial to be printed
// MODIFIES: Poly -- Assign numbers to letters in the polynomial
void AssignNumbersToLetters (polynomial* Poly)
{
  term_head* temp_head;
  int Current_letter_numbers[NUM_LETTERS];
  unsigned int i;

  if (Poly == NULL)
	{
	  std::cout << "AssignNumbersToPoly(Poly): Poly is null pointer" 
				<< std::endl;
	}
  else 
	{
	  temp_head = Poly->terms;
	  while (temp_head != NULL) 
		{
		  for (i=0; i<NUM_LETTERS; i++)
			{
			  Current_letter_numbers[i] = 1;
			}
		  AssignNumbersToTerm (temp_head->term, Current_letter_numbers);
		  temp_head = temp_head->next;
		}
    }
}



void AssignNumbersToTerm (term_node* Pntr, int Cln[])
{
  if (((Pntr->left) == NULL) && ((Pntr->right) == NULL))
	{
	  Pntr->number = Cln[Pntr->letter - 'a']++;
	}
  else 
	{
	  AssignNumbersToTerm (Pntr->left,  Cln);
	  AssignNumbersToTerm (Pntr->right, Cln);
    }
}



void DestroyPoly (struct polynomial* Poly)
{
  assert_not_null_retvoid(Poly);
  DestroyTerms (Poly->terms);
  free (Poly);
}


void DestroyTerms (struct term_head* Term_head)
{
  assert_not_null_retvoid(Term_head);
  if (Term_head->next == NULL) 
	{
	  FreeNodes (Term_head->term);
	  free (Term_head);
    }
  else 
	{
	  DestroyTerms (Term_head->next);
	  FreeNodes (Term_head->term);
	  free (Term_head);
    }
}


void FreeNodes (struct term_node* Term_node)
{
  assert_not_null_retvoid(Term_node);
  if ((Term_node->left == NULL) && (Term_node->right == NULL))
	{
	  free (Term_node);
	}
  else 
	{
	  FreeNodes (Term_node->left);
	  FreeNodes (Term_node->right);
	  free (Term_node);
    }
}


// end of Po_routines.cc
