/*
  file:         Type_table.cc
  created:      David P Jacobs and Sekhar Muddana May 1990
  modified:     pasha jun 19,29 2019
  modification: some c structures converted to std::vector's
*/

// This module contains routines dealing with Type table.
// Given the Target_type, generate a Type_table consisting of all the subtypes 
// of the Target_type. Also generate data structures to map Degree to subtype 
// and subtype to Type_table.                                 
// NOTE: Target_type[] is assumed to end with a 0 in the last position.

#include <stdlib.h>
#include <iostream>
#include "albert.hh"

#include "Type_table.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

bool FillTypecount (const unsigned int);
void FillTypetable (const unsigned int);
int  GetIndex (const Type&);
bool InitStoreblocksizes();
int  InitTypetable();

static Type Target_type_Type_table;          // Input from higher level module
static unsigned int Target_type_len = 0;     // Computed from Target_type
static int  Target_type_deg_Type_table = 0;  // Computed from Target_type
static std::vector<int> Type_count;          // Used to fill Type_table
static std::vector<TT_node> Type_table;      // Heart of the matter
static std::vector<int> Type_table_index;         // Map type to Type_table

// Computed from Type_count. Size of Type_table
static unsigned int Tot_subtypes = 0;         

static std::vector<int> Deg_to_type_table_index; // Map Degree to Type_table
static std::vector<int> Temp_dttt_index;       // Used to fill Type_table_index
static std::vector<int> Store_block_sizes;     // To find offset into Type_table



// Compute all subtypes of the given target type.
// Allocate memory for all these subtypes i.e for Type_table[].
// Initialize Type_table_index[] -- Type to Type_table.
// Initialize Deg_to_type_table_index[] -- Degree to Type.
// GLOBALS INITIALIZED:                                           
//     Target_type_len -- Length of the Target_type.              
//     Target_type_deg -- Degree of the Target_type.               
//     Type_count[i] -- Num. of i'th Degree subtypes in Target_type.
//     Type_table[] -- Array of all the subtypes in ascending Degree.
//     Type_table_index[] -- Offsets into Type_table[].             
//     Tot_subtypes -- Size of the Type_table.                     
//     Store_block_sizes[] -- Multiplication constants to find Index.
// arg: Cur_type[] -- to build the Type_table[]
// RETURNS: 1 if Type_table[] is successfully built, 0 otherwise
bool CreateTypeTable (const Type& Cur_type)
{
  Target_type_len = 0;
  unsigned int i=0;
  while (Cur_type[i++] != 0)
	{
	  Target_type_len++;
	}
  
  GetNewType (Target_type_Type_table);
  for (i=0; i<Target_type_len; i++)
	{
	  Target_type_Type_table[i] = Cur_type[i];
	}
  Target_type_deg_Type_table = GetDegree (Target_type_Type_table); 

  if (InitStoreblocksizes() != 1)
	{
	  return (0);
	}

  Type_count.clear();
  for (int i=0; i <= Target_type_deg_Type_table; i++)
	{
	  Type_count.push_back (0);
	}
    
  FillTypecount (0);

  if (InitTypetable() != 1)
	{
	  return (0);
	}

  Temp_dttt_index.clear();
  Type_count.clear();
  return (1);
} // end of CreateTypeTable()



unsigned int GetTargetLen()
{
  return (Target_type_len);
}


void GetNewType (Type& t)
{
  t.resize (Target_type_len + 1);
  for (unsigned int i=0; i <= Target_type_len; i++)
	{
	  t[i] = 0;
	}
}


void NameToType (const Name N, Type& T)
{
  for (unsigned int i=0; i<Target_type_len; i++)
	{
	  T[i] = Type_table[N].type[i];
	}
}


void SubtractTypeName (const Name n1, const Name n2, Name* res_name)
{
  Type temp_type;
  GetNewType (temp_type);
  for (unsigned int i=0; i<Target_type_len; i++)
	{
	  temp_type[i] = Type_table[n1].type[i] - Type_table[n2].type[i];
	}
  *res_name = TypeToName (temp_type);
}


int GetDegree (const Type& Pntr)
{
  int deg = 0;
  for (unsigned int i=0; i<Target_type_len; i++)
	{
	  deg += Pntr[i];
	}
  return (deg);
}


int GetDegreeName (const Name n)
{
  return (GetDegree (Type_table[n].type));
}


Name TypeToName (const Type& T)
{
  return (Type_table_index[GetIndex(T)]);
}


bool IsSubtype (const Name n1, const Name n2)
{
  bool is_sub_type = 1;
  for (unsigned int i=0; i<Target_type_len; i++)
	{
	  if (Type_table[n1].type[i] > Type_table[n2].type[i])
		{
		  is_sub_type = 0;
		}
	}
  return (is_sub_type);
}


void EnterBeginBasis (const int TTindex, const Basis basis)
{
  Type_table[TTindex].begin_basis = basis;
}


void EnterEndBasis (const int TTindex, const Basis basis)
{
  Type_table[TTindex].end_basis = basis;
}

void UpdateTypeTable 
(const Name n, const Basis Begin_basis, const Basis End_basis)
{
  EnterBeginBasis (n, Begin_basis);
  EnterEndBasis   (n, End_basis);
}


bool SameType (const Type T1, const Type T2)
{
  for (unsigned int i=0; i<Target_type_len; i++)
	{
	  if (T1[i] != T2[i])
		{
		  return (0);
		}
	}
  return(1);
}



// Fill the Store_block_sizes[].
// GLOBALS FILLED: 
//     Store_block_sizes[] -- Multiplication constants to get Index.
// GLOBALS REQUIRED: Target_len -- Length of the Target_type.
// RETURNS: 1 if successfull, 0 otherwise
bool InitStoreblocksizes()
{
  Store_block_sizes.resize (Target_type_len);
  Store_block_sizes[Target_type_len - 1] = 1;
  for (int i = Target_type_len - 2; i>=0; i--)
	{
	  Store_block_sizes[i] = 
		(Target_type_Type_table[i+1] + 1) * Store_block_sizes[i+1]; 
	}
  // PrintSbsizes();
  return (1);
}





// Count all subtypes of the given Target_type
// GLOBALS MODIFIED:                                              
//     Type_count[i] -- num.of subtypes of Degree i in Target_type.
//     Target_type[] -- is modified on each call, but restored.   
// GLOBALS REQUIRED:                                              
//     Target_len -- Length of the Target_type.                    
// arg: Cur_scan_pos -- Current scanner position in the Target_type.
// RETURNS: 1 if successfull, 0 otherwise
bool FillTypecount (const unsigned int Cur_scan_pos)
{
  if (Cur_scan_pos > Target_type_len) 
	{    
	  int d = 0;
	  for (unsigned int i=0; i<Target_type_len; i++)
		{
		  d += Target_type_Type_table[i];
		}
	  Type_count[d]++;
    }
  else 
	{
	  int save;
	  for (int i=0; i <= Target_type_Type_table[Cur_scan_pos]; i++) 
		{
		  save = Target_type_Type_table[Cur_scan_pos];
		  Target_type_Type_table[Cur_scan_pos] = i;
		  FillTypecount (Cur_scan_pos + 1);
		  Target_type_Type_table[Cur_scan_pos] = save;
        }
    }
  return (1);
} // end of FillTypecount()




// Allocate memory for all these subtypes i.e for Type_table[].
// Initialize Type_table_index[] -- Type to Type_table.
// Initialize Deg_to_type_table_index[] -- Degree to Type.
// GLOBALS INITIALIZED:                                         
//     Type_table[] -- Array of all the subtypes in ascending Degree.
//     Type_table_index[] -- Offsets into Type_table[].             
//     Tot_subtypes -- Size of the Type_table.                     
// REQUIRES:                                                       
//     Type_count[i] -- Num. of i'th Degree subtypes in Target_type.*/
// RETURNS: 1 if Type_table[] is successfully built, 0 otherwise
int InitTypetable()
{
  Tot_subtypes = 0;
  for (int i=0; i <= Target_type_deg_Type_table; i++)
	{
	  Tot_subtypes += Type_count[i];
	}

  Type_table.resize (Tot_subtypes);

  for (unsigned int i=0; i<Tot_subtypes; i++) 
	{
	  GetNewType (Type_table[i].type);
	  Type_table[i].begin_basis = Type_table[i].end_basis = 0;
    }
   
  Deg_to_type_table_index.clear();
  Deg_to_type_table_index.push_back (0);
  for (int i=1; i <= Target_type_deg_Type_table; i++)
	{
	  Deg_to_type_table_index.push_back 
		(Deg_to_type_table_index[i-1] + Type_count[i-1]);
	}

  Temp_dttt_index.clear();
  for (int i=0; i <= Target_type_deg_Type_table; i++)
	{
	  Temp_dttt_index.push_back (Deg_to_type_table_index[i]);
	}
  
  Type_table_index.resize (Tot_subtypes);
  FillTypetable (0);
  return (1);
} // end of InitTypetable()



void DestroyTypeTable()
{
  Deg_to_type_table_index.clear();
  Type_table_index.clear();
  Store_block_sizes.clear();
}



// Fill the Type_table.                               
// GLOBALS FILLED:                                           
//     Type_table[] -- All the subtypes.                     
//     Type_table_index[] -- Offsets into Type_table.             
// GLOBALS REQUIRED:                                              
//     Target_len -- Length of the Target_type.                   
// GLOBALS MODIFIED:                                              
//     Temp_dtttx_index[] -- Where to place the next subtype.     
// arg: Cur_scan_pos -- Current scanner position in the Target_type.
void FillTypetable (unsigned int Cur_scan_pos)
{
  if (Cur_scan_pos > Target_type_len) 
	{    
	  unsigned int i;
	  int d=0;
	  for (i=0; i<Target_type_len; i++)
		{
		  d += Target_type_Type_table[i];
		}
	  for (i=0; i<=Target_type_len; i++)
		{
		  Type_table[Temp_dttt_index[d]].type[i] = Target_type_Type_table[i];
		}
	  Type_table_index[GetIndex(Target_type_Type_table)] = Temp_dttt_index[d];
	  Temp_dttt_index[d]++;
    }
  else 
	{
	  int save;
	  for (int i=0; i <= Target_type_Type_table[Cur_scan_pos]; i++) 
		{
		  save = Target_type_Type_table[Cur_scan_pos];
		  Target_type_Type_table[Cur_scan_pos] = i;
		  FillTypetable (Cur_scan_pos + 1);
		  Target_type_Type_table[Cur_scan_pos] = save;
        }
    }
}



// GLOBALS REQUIRED:                              
//     Target_len -- Length of the Target_type.   
//     Store_block_sizes[] -- Multiplication constants to get Index.
// arg: Type for which index into Type_table[] is to be found
// RETURNS: Index into the Type_table
int GetIndex (const Type& Pntr)
{
  int result = 0;
  for (unsigned int i=0; i<Target_type_len; i++) 
	{
	  result += Pntr[i] * Store_block_sizes[i];
	}
  return (result);
}


// RETURNS: pointer to first subtype in Type_table[]
Name FirstTypeDegree (const Degree D)
{
  if ((D >= 0) && (D <= Target_type_deg_Type_table))
	{
	  return (Deg_to_type_table_index[(unsigned int)D]);
	}
  else
	{
	  return (-1);
	}
}


// arg: Pntr -- pointer to a subtype in a TT_node in Type_table[]
// RETURNS: pointer to next subtype in TT_node in Type_table[]
Name NextTypeSameDegree (const Name n)
{
  if (n >= ((int)Tot_subtypes - 1))
	{
	  return (-1);
	}
  if (GetDegreeName(n + 1) > GetDegreeName(n))
	{
	  return (-1);
	}
  return (n+1);
}


Basis FirstBasis (const Name N)
{
  return (BeginBasis (N));
}


Basis Albert::NextBasisSameType (const Basis B) const
{
  if ((unsigned int)(B+1) >= this->Basis_table.size())
	{
	  return (0);
	}
  if (this->GetType (B) != this->GetType (B+1))	 // B,B+1 valid basis numbers
	{
	  return (0);	 // but have different types
	}
  else
	{
	  return (B+1);	 // Have same types
	}
}



// arg: Pntr -- pointer to a subtype in a TT_node in Type_table[]
// RETURNS: First Basis element with the given type
Basis BeginBasis (const Name n)
{
  return (Type_table[n].begin_basis);
}


// arg: Pntr -- pointer to a TT_node in Type_table[]
// RETURNS: Last Basis element with the given type
Basis EndBasis (const Name n)
{
  return (Type_table[n].end_basis);
}


int NumberOfBasis (const Name n)
{
  return (Type_table[n].end_basis - Type_table[n].begin_basis + 1);
}


void PrintType (const Type& Pntr, std::ostream& os)
{
  for (unsigned int i=0; i<Target_type_len; i++)
	{
	  os << (int)Pntr[i];
	}
}


void PrintTypeName (const Name n, std::ostream& os)
{
  PrintType (Type_table[n].type, os);
}


void PrintTypetable()
{
  std::cout << "Type Table: " << std::endl;
  unsigned int j;
  for (unsigned int i=0; i<Tot_subtypes; i++) 
	{
	  std::cout << "   ";
	  for (j=0; j<Target_type_len; j++)
		{
		  std::cout << Type_table[i].type[j];
		}
	  std::cout << " " << std::setw(3) << Type_table[i].begin_basis 
				<< " " << std::setw(3) << Type_table[i].end_basis << std::endl;
    }
}


void PrintTypetableindex()
{
  std::cout << "Type table index entries are :" << std::endl;
  for (unsigned int i=0; i<Tot_subtypes; i++)
	{
	  std::cout << Type_table_index[i] << std::endl;
	}
}


// end of Type_table.cc
