/*
  file:         albert.cc
  purpose:      Albert class
  created:      pasha mar 10 2009
  modified:     pasha may 18-19,jun 16,18-19,22,26-29 2019
  modification: - reading config file transfered from ctor to a member func
                - set_char -> set_field
				- eliminated clear_mult_table()
				- some char* converted to std::string
				- started to add support for number fields via LiDIA
*/

#include <gmpxx.h>
#include <string.h>
#include <unistd.h>
#include <fstream>
#include <LiDIA/polynomial.h>
#include "albert.hh"
#include "GF2.hh"
#include "GFp.hh"
#include "lidia_alg_num.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#define MAXPATH 200

// built-in modular arithemtic
unsigned long long Albert::PRIME;
unsigned long long GFp_element::HALF_PRIME;
LiDIA::number_field Albert::NF;
std::map<unsigned long long, long long> GFp_element::Inverse_table;
std::map<unsigned long long, long long>::const_iterator GFp_element::ITI;


Albert::Albert ()
{
  this->block_list_head = NULL;
  this->current_block   = NULL;
  this->mtable_present  = 0;
  this->mult_table      = NULL;
  this->node_list_head  = NULL;
  this->status_GenerateEquations = 1;
  this->status_Multpart          = 1;
  this->status_PerformSubs       = 1;
  this->Deg_to_basis_table.clear();

  this->field_elem_type = INVALID_TYPE;

  // by default, set field to GF(251)
  this->set_field ("251", 0);

  this->Id_queue.first = NULL;
}



Albert::~Albert()
{
  /*
  for (std::list<dalbert_node>::iterator it = this->Dalbertheadptr.begin(); 
	   it != this->Dalbertheadptr.end(); ++it)
	{
	}
  */
 
  this->destroy_mult_table();
} // end of Albert dtor




////////////////////////////////////////////////////////////////////////////
//                       read_config_file                                 //
////////////////////////////////////////////////////////////////////////////
void Albert::read_config_file (const char* albert_dot_file, 
							   const bool verbose)
{
  // Fill in Dalbertheadptr args with the structure from .albert file, 
  // (empty if .albert not present or unable to read).
  // Reads the .albert file if present and forms a list of definitions.
  // Linked list representation means there could be any number of definitions 
  // in .albert.
  // A definition can start anywhere on a line.
  // Blank lines are permitted anywhere in .albert.
  // Line after % is totally ignored.
  // A definition can be continued on the next line by ending it with a /.
  // Comments can appear between two lines of a definition or at the of any 
  // line.
  // A blank ' ' or a '\t' seperates the left hand side and right hand side of a
  // definition.
  // If Albert dumps the core during initialization, probably there is a bug in
  // this code. Albert can be used either by changing the .albert file or by 
  // getting rid of it completely.

  std::ifstream albert_fp;
  albert_fp.open (albert_dot_file);
  if (! albert_fp.is_open())
    {
  	  if (verbose)
		{
		  std::cout << "Can't open " << albert_dot_file << " file." 
					<< std::endl;
		}
	  return;
    }
  else if (verbose)
    {
	  char* fullpath = realpath (albert_dot_file, NULL);
      std::cout << "Using " << fullpath << std::endl;
	  free (fullpath);
    }

  std::string albert_line;
  std::string temp_line;
  std::string temp_line2;
  bool temp;  // flag to indicate wheter eof occured
  int temp_line_len, lhs_len, rhs_len, alinelen, lhsstart;
  int i=0;
  bool slashfound = 1;
  bool blankfound = 1;
  while (std::getline (albert_fp, temp_line), temp = albert_fp.good())
	{
	  dalbert_node dan;
	  dan.lhs.clear();
	  dan.rhs.clear();

	  slashfound = 1;
	  while (slashfound && temp)
		{
		  blankfound = 1;
		  while (blankfound && temp)
			{
			  temp_line_len = temp_line.length();
			  for (i=0; i<temp_line_len; i++) 
				{
				  if (temp_line[i] == '%') 
					{
					  temp_line[i] = '\0';
					  break;
					}
				}
			  temp_line_len = temp_line.length();
			  for (i=0; i<temp_line_len; i++) 
				{
				  if ((temp_line[i] != ' ') && (temp_line[i] != '\t') &&
					  (temp_line[i] != '\n'))
					{
					  blankfound = 0;
					}
				}
			  if (blankfound)
				{
				  std::getline (albert_fp, temp_line);
				  temp = albert_fp.good();
				}
			}
		  i = 0;
		  temp_line2.resize (0);
		  while ((temp_line[i] != '\\') && (temp_line[i] != '\0') &&
				 (temp_line[i] != '%') && (temp_line[i] != '\n'))
			{
			  temp_line2.push_back (temp_line[i++]);
			}
		  if (temp_line[i] == '\\') 
			{
			  slashfound = 1;
			  std::getline (albert_fp, temp_line); temp = albert_fp.good();
			}
		  else
			{
			  slashfound = 0;
			}
		  albert_line += temp_line2;
		}
	  
	  i = 0;
	  albert_line.push_back ('\0');
	  alinelen = albert_line.length();
	  while (i < alinelen) 
		{
		  if ((albert_line[i] != ' ') && (albert_line[i] != '\t'))
			{
			  break;
			}
		  i++;
	  }
	  lhs_len = 0;
	  lhsstart = i;
	  while (i < alinelen) 
		{
		  if ((albert_line[i] == ' ') || (albert_line[i] == '\t'))
			{
			  break;
			}
		  lhs_len++;
		  i++;
		}
	  rhs_len = 0;
	  while (i < alinelen) 
		{
		  if ((albert_line[i] != ' ') && (albert_line[i] != '\t'))
			{
			  rhs_len++;
			}
		  i++;
		}
	  if ((lhs_len > 0) && (rhs_len > 0))  
		{
		  i = lhsstart;
		  while (i < alinelen) 
			{
			  if ((albert_line[i] == ' ') || (albert_line[i] == '\t'))
				{
				  break;
				}
			  dan.lhs.push_back (albert_line[i++]);
			}

		  while (i < alinelen) 
			{
			  if ((albert_line[i] != ' ') && (albert_line[i] != '\t'))
				{
				  dan.rhs.push_back (albert_line[i]);
				}
			  i++;
			}
		  this->Dalbertheadptr.push_back (dan);
		}
	  albert_line.resize (0);
	} // end of loop thru lines of .albert file

  if (albert_fp.is_open())
	{
	  albert_fp.close();
	}

#if DB_READ_ALBERT
  i = 1;
  for (std::list<dalbert_node>::iterator it = this->Dalbertheadptr.begin(); 
	   it != this->Dalbertheadptr.end(); ++it)
	{
      std::cout << " " << i++ << " " 
				<< it->lhs.size() << "  " << it->lhs << " " 
				<< it->rhs.size() << "  " << it->rhs << " " 
				<< std::endl;
	}
#endif
  
} // end of Albert::read_config_file()




////////////////////////////////////////////////////////////////////////////
//                       destroy_mult_table                               //
////////////////////////////////////////////////////////////////////////////
void Albert::destroy_mult_table()
{
  switch (this->field_elem_type)
	{
	case GF2_TYPE:
	  this->DestroyMultTable<GF2_element>();
	  break;
		  
	case GFP_TYPE:
	  this->DestroyMultTable<GFp_element>();
	  break;
	  
	case GMP_TYPE:
	  this->DestroyMultTable<mpq_class>();
	  break;
	  
	case LIDIA_TYPE:
	  this->DestroyMultTable<alg_number>();
	  break;

	case INVALID_TYPE:
	  return;

	default:
	  std::cerr << "wrong field_elem_type=" << this->field_elem_type
				<< std::endl;
	  break;
	}
  this->mult_table = NULL;
  if (this->mtable_present) 
	{
	  this->mtable_present = 0;
	  std::cout << "Destroyed the Multiplication Table." << std::endl;
	}
} // end of Albert::destroy_mult_table()




////////////////////////////////////////////////////////////////////////////
//                       print_field                                    //
////////////////////////////////////////////////////////////////////////////
// arg: 
// output_for_gap - whether to format output for GAP (1) or for human (0)
void Albert::print_field (std::ostream& os, const bool output_for_gap) const
{
  switch (this->field_elem_type)
	{
	case GF2_TYPE:
	  os << "GF(2)";
	  break;

	case GFP_TYPE:
	  os << "GF(" << Albert::PRIME << ")";  
	  break;
	  
	case GMP_TYPE:
	  os << (output_for_gap ? "Rationals" : "Q");
	  break;

	case LIDIA_TYPE:
	  if (output_for_gap)
		{
		  // this "F" is the algebraic number field Albert::NF defined
		  // in GAP terms; see Print_MultTable()
		  os << "F";
		}
	  else
		{
		  os << "Q(" << Albert::NF << ")";
		}
	  break;

	default:
	  std::cerr << "wrong field_elem_type=" << this->field_elem_type
				<< std::endl;
	  break;
	}
} // end of Albert::print_field()




////////////////////////////////////////////////////////////////////////////
//                       print_mtable                                     //
////////////////////////////////////////////////////////////////////////////
void Albert::print_mtable (std::ostream& os, const bool output_for_gap) const
{
  switch (this->field_elem_type)
	{
	case GF2_TYPE:
	  this->Print_MultTable<GF2_element> (os, output_for_gap);
	  break;

	case GFP_TYPE:
	  this->Print_MultTable<GFp_element> (os, output_for_gap);
	  break;
	  
	case GMP_TYPE:
	  this->Print_MultTable<mpq_class> (os, output_for_gap);
	  break;

	case LIDIA_TYPE:
	  this->Print_MultTable<alg_number> (os, output_for_gap);
	  break;

	default:
	  std::cerr << "wrong field_elem_type=" << this->field_elem_type
				<< std::endl;
	  break;
	}
} // end of Albert::print_mtable()





////////////////////////////////////////////////////////////////////////////
//                       set_field                                        //
////////////////////////////////////////////////////////////////////////////
// set charactersitic of the ground field
// returns 1 on success, 0 on failure
bool Albert::set_field (const std::string& f, const bool verbose)
{
  // the string f can be one of the following:
  // "Q": rationals
  // number: GF(number)
  // Q(polynomial) (e.g. Q(x^2 + x + 1)): number field Q(polynomial)
  if (f.compare ("Q") == 0)
    {
      // rational arithmetic via GMP
	  if (this->field_elem_type != GMP_TYPE)
		{
		  this->change_field (GMP_TYPE, verbose);
		}
    }
  else if 
	((f.compare (0, 2, "Q(") == 0) && (f.compare (f.size()-1, 1, ")") == 0))
	{
	  // number fields via LiDIA
	  // poly supposed to be a polynomial (like x^2 + x + 1)
	  std::istringstream iss (f.substr (2, f.size()-3));
	  LiDIA::polynomial<LiDIA::bigint> lidia_poly;
	  iss >> lidia_poly;

	  // check that the polynomial is monic
	  if (lidia_poly.lead_coeff() != 1)
		{
		  std::cout << "polynomial " << lidia_poly << " is not monic" 
					<< std::endl;
		  return (0);
		}

	  LiDIA::number_field new_nf = LiDIA::number_field (lidia_poly);
	  /*
	  if (this->field_elem_type == LIDIA_TYPE)
		{
		  if (Albert::NF != new_nf)
			{
			  this->change_number_field (new_nf, verbose);
			}
		}
	  else
		{
		  this->change_number_field (new_nf, verbose);
		}
	  */
	  if ((this->field_elem_type != LIDIA_TYPE) || (Albert::NF != new_nf))
		{
		  this->change_number_field (new_nf, verbose);
		}
	}
  else 
	{
	  // supposed to be the characteristic of the ground field
	  long long int new_char = atoll (f.c_str());
	  if (new_char < 2)
		{
		  std::cerr << "wrong value of characteristic: " << new_char
					<< ". Ground field not changed." << std::endl;
		  return (0);
		}
	  else if (new_char == 2)
		{
		  if (this->field_elem_type != GF2_TYPE)
			{
			  this->change_field (GF2_TYPE, verbose);
			}
		}
	  else
		{
		  // built-in modular arithmetic
		  // check that new_char is a prime
		  char str[20];
		  sprintf (str, "%lld", new_char);
		  mpz_t I;
		  mpz_init (I);
		  mpz_set_str (I, str, 10);
		  int r = mpz_probab_prime_p (I, 50);
		  mpz_clear (I);
		  switch (r)
			{
			case 0:
			  std::cout << new_char << " is not prime" << std::endl;
			  return (0);
			case 1:
			  std::cout << new_char << " is probably prime" << std::endl;
			  break;
			case 2:
			  // new_char is prime
			  break;
			default:
			  std::cerr << "wrong return value " << r 
						<< " when testing for primeness" << std::endl;
			  return (0);
			}

		  /*
		  if (this->field_elem_type == GFP_TYPE)
			{
			  if (Albert::PRIME != (unsigned long long)new_char)
				{
				  this->change_prime_field (new_char, verbose);
				}
			}
		  else
			{
			  this->change_prime_field (new_char, verbose);
			}
		  */
		  if ((this->field_elem_type != GFP_TYPE) || (Albert::PRIME != (unsigned long long)new_char))
			{
			  this->change_prime_field (new_char, verbose);
			}
		}
	}

  return (1);
} // end of Albert::set_field()



void Albert::change_field (const FieldElemType elem_type, const bool verbose)
{
  this->destroy_mult_table();
  this->field_elem_type = elem_type;
  if (verbose)
	{
	  std::cout << "Ground field changed to ";
	  this->print_field (std::cout, 0);
	  std::cout << std::endl;
	}
}

void Albert::change_prime_field (const long long new_char, const bool verbose)
{
  Albert::PRIME = (unsigned long long)new_char;
  GFp_element::HALF_PRIME = (new_char - 1) >> 1;
  GFp_element::Inverse_table.clear();
  this->change_field (GFP_TYPE, verbose);
}

void Albert::change_number_field (const LiDIA::number_field& nf, 
								  const bool verbose)
{
  this->destroy_mult_table();
  this->field_elem_type = LIDIA_TYPE;
  Albert::NF = nf;
  if (verbose)
	{
	  std::cout << "Ground field changed to ";
	  this->print_field (std::cout, 0);
	  std::cout << std::endl;
	}
}


// end of albert.cc
