/*
  file:         albert.hh
  purpose:      Albert class
  created:      pasha feb 24,27-28,mar 1-3,5,7-10,12-14,may 1,17-19,22 2016
  modified:     pasha may 19,jun 16,18-19,22,27-29 2019
  modification: - reading config file transfered from ctor to a member func
				- started to add support for number fields via LiDIA
				- some methods eliminated (called once)
				- some c structures turned to c++ containers
				- eliminated Deg_var_types and Deg_vars
*/

#ifndef __ALBERT_H
#define __ALBERT_H

#include <gmpxx.h>
#include <iomanip>
#include <iostream>
#include <list>
#include <map>
#include <set>
#include <LiDIA/alg_number.h>
#include "assert_not_null.hh"
#include "dotalbert.hh"
#include "mpq.hh"
#include "node_mgt.hh"
#include "Alg_elements.hh"
#include "Basis_table.hh"
#include "Build.hh"
#include "CreateMatrix.hh"
#include "GenerateEquations.hh"
#include "Generators.hh"
#include "Id_routines.hh"
#include "Memory_routines.hh"
#include "Mult_table.hh"
#include "Mult_table_print.hh"
#include "PerformSub.hh"
#include "Po_create_poly.hh"
#include "Po_parse_exptext.hh"
#include "Po_routines.hh"
#include "SparseReduceMatrix.hh"
#include "Type_table.hh"


enum FieldElemType
  {
	GFP_TYPE,      // built-in implementation for modular arithmetic
	GF2_TYPE,
	GMP_TYPE,      // for rational arithmetic
	LIDIA_TYPE,    // number fields via LiDIA
	INVALID_TYPE
  };



class Albert
{
public:

  FieldElemType       field_elem_type;
  bool                mtable_present;
  std::vector<BT_rec> Basis_table;
  polynomial*         Cur_id;
  size_t              Cur_poly_len;  // length of the expanded Cur_poly string
  std::list<dalbert_node> Dalbertheadptr;
  std::vector<Deg_to_basis_rec> Deg_to_basis_table;
  id_queue_head       Id_queue;      // head to Queue of identities
  Type                Target_type;
           
  // variables for measuring density of last generated matrix
  bool gather_density_flag;
  unsigned long int num_elements;
  unsigned long int max_num_elements;
  unsigned long int matrix_size;

  // characteristic of the ground field
  // in use only when field elements are of GFP_TYPE
  static unsigned long long PRIME;  

  // current number field; in use only when field elements are of LIDIA_TYPE
  static LiDIA::number_field NF;    

  Albert();
  ~Albert();
  
  void destroy_mult_table();
  void print_field  (std::ostream&, const bool) const;
  void print_mtable (std::ostream&, const bool) const;
  void read_config_file (const char*, const bool);
  bool set_field (const std::string&, const bool);
  void PrintBasisTable (std::ostream&) const;



  ////////////////////////////////////////////////////////////////////////////
  //                       add_identity                                     //
  ////////////////////////////////////////////////////////////////////////////
  // returns: 
  // (positive) number of identity in the list of all identities on success, 
  // 0 in failure
  unsigned short add_identity (const char* identity)
  {
	this->Cur_id = Create_poly (identity, &this->Cur_poly_len);
	if (this->Cur_id == NULL)
	  {
		std::cerr << "albert: " << identity << " is not a polynomial" 
				  << std::endl;
		return (0);
	  }
	if (! Homogeneous (this->Cur_id)) 
	  {
		std::cerr << "albert: identity " << identity << " is not homogeneous.";
		return (0);
	  }
	if (this->Cur_id->degree < 2) 
	  {
		std::cerr << "albert: total degree of identity " << identity 
				  << " must be at least 2" << std::endl;
		return (0);
	  }
	AssignNumbersToLetters (this->Cur_id);
	const unsigned short Cur_id_no = 
	  Add_id (this->Cur_id, identity, &this->Id_queue);
	this->destroy_mult_table();
	return (Cur_id_no);
  } // end of Albert::add_identity()






  ////////////////////////////////////////////////////////////////////////////
  //                       build                                            //
  ////////////////////////////////////////////////////////////////////////////
  // This is the highest level function in this part of the system. It gets 
  // called from the command interpreter when "build" is invoked.
  // For each degree, New Basis are created and New Products are entered.
  // MODIFIES:                                                     
  //     mult_table -- is fully built.                             
  //     Basis_table -- will be Complete.                          
  // REQUIRES:                                                      
  //     Target_Type -- Target to be reached.                       
  //     Identities -- All the identities entered by User.          
  // returns 1 on success, 0 on failure
  template<typename T> bool build (const bool verbose)
  {
	this->DestroyMultTable<T>();
	this->CreateMultTable<T>();
	if (verbose)
	  {
		std::cout << "Building the Multiplication Table." << std::endl;
	  }
	time_t Start_time;
	time (&Start_time);
	if (verbose)
	  {
		std::cout << std::endl << "Build begun at " 
				  << ctime (&Start_time) << std::endl
				  << "Degree    Current Dimension   Elapsed Time(in seconds)" 
				  << std::endl;
	  }

	this->mtable_present = CreateTypeTable (this->Target_type);

	const int Target_degree = GetDegreeName (TypeToName (this->Target_type));
	int i;
	if (this->mtable_present) 
	  { 
		// Initialize the Basis Table (to zeroes) and Nextbasistobefilled (to 1)
		this->Basis_table.clear();
		  
		// first entry is expected to be zero
		BT_rec br;
		br.left_factor  = 0;
		br.right_factor = 0;
		br.type = 0;
		this->Basis_table.push_back (br);
		this->Deg_to_basis_table.clear();

		for (i=0; i < Target_degree; i++) 
		  {
			Deg_to_basis_rec d;
			d.first_basis = 0;
			d.last_basis  = 0;
			this->Deg_to_basis_table.push_back (d);
		  }

		this->Deg_of_last_basis = 0;
		this->mtable_present = 1;
	  }

	this->gather_density_flag = 0;	
	if (this->mtable_present)
	  {
		time_t Current_time;
		for (i=1; i <= Target_degree; i++)  
		  {
			if (i == Target_degree)
			  {
				this->gather_density_flag = 1;
			  }
			this->mtable_present = this->ProcessDegree<T> (i);
			if (! this->mtable_present)
			  {
				break;
			  }
			time (&Current_time);
			std::cout << "  " 
					  << std::setw(2) << i 
					  << "             " 
					  << std::setw(5) << this->Current_dimension 
					  << "             " 
					  << std::setw(5) << (Current_time - Start_time)
					  << std::endl;
		  }
	  }
#if PRINT_BASIS_TABLE
	PrintBasisTable();
#endif
#if PRINT_MULT_TABLE
	Print_MultTable();
#endif
	
	DestroyTypeTable();

	if (verbose)
	  {
		if (this->mtable_present)
		  {
			std::cout << "Build completed. " << std::endl;
			const double density = 100 * 
			  (double)this->max_num_elements/(double)this->matrix_size ;
			if (density > 0.0)
			  {
				_print_density (density);
			  }
			else
			  {
				if (this->matrix_size == 0)
				  {
					std::cout << "No Matrix." << std::endl;
				  }
				else
				  {
					_print_density (density);
				  }
			  }
		  }
		else
		  {
			std::cout << "Build incomplete" << std::endl;
		  }
	  }
	
	return (this->mtable_present);
  } // end of Albert::build()





  ////////////////////////////////////////////////////////////////////////////
  //                       IsIdentity                                       //
  ////////////////////////////////////////////////////////////////////////////
  // Called from the Main(), when polynomial command is issued.
  // Polynomial is expanded using the multiplication table.
  // If the expansion collapses to 0, that means Poly is an identity.
  template<typename T> bool IsIdentity (const polynomial* Poly) const
  {
	assert_not_null(Poly);
	const term_head* temp_head = Poly->terms;
	typename Alg_element<T>::type result;
	bool status = 1;
	while (temp_head != NULL) 
	  {
		typename Alg_element<T>::type ae;
		this->ExpandTerm<T> (ae, temp_head->term, &status);
		if (! status)
		  {
			std::cout << "Severe Bug. Cant ExpandTerm. Basis product undefined"
					  << std::endl;
			return (0);
		  }
		ScalarMultAE<T> ((T)(temp_head->coef), ae);
		this->AddAE<T> (ae, result);
		temp_head = temp_head->next;
	  }
	return (IsZeroAE<T> (result));
  } // end of Albert::IsIdentity()




private:

  // Here and on many other occasions below: we cannot have templated class 
  // members, thus declaring them as void* and performing casts as necessary;
  // should be casted to mem_block<T>*
  void*                 block_list_head; 

  void*                 current_block;   // should be casted to mem_block<T>*

  // should be casted to 
  // std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*
  void*                 mult_table;

  void*                 node_list_head;  // should be casted to Node<T>*

  // should be casted to Basis_pair_node<T>*
  void*                 running_list;   

  bool                  status_GenerateEquations;
  bool                  status_Multpart;
  int                   status_PerformSubs;
  std::vector<Basis>    BasisNames;
  Unique_basis_pair_list ColtoBP;   // from ExtractMatrix.cc
  Unique_basis_pair_list ColtoBP_CreateMatrix;
  std::vector<int>      Cur_deg_vars;
  std::vector<int>      Cur_index_var;
  Type                  Cur_type;
  int                   Cur_type_degree;
  Basis                 Current_dimension;
  int                   Deg_of_last_basis;
  std::vector<int>      Deg_var;
  std::vector<int>      Dependent;
  
  // should be casted to Eqn_list_node<T>*
  void*                 First_eqn_list_node;

  // should be casted to Basis_pair_node<T>*
  void*                 Local_list; 

  void**                Matrix_Base_Ptr; // should be casted to Node<T>**
  unsigned int          MatrixRank;
  int                   Max_deg_var;

  Name                  N_CreateMatrix;
  unsigned int          Num_cols_ExtractMatrix;
  int                   Num_equations;
  int                   Num_rows_ExtractMatrix;
  int                   Num_vars;
  std::set<std::pair<Basis, Basis> > Pair_present;
  std::vector<Perm>     Permutation_list;
  Type                  Seq_sub_types;
  std::vector<Name>     Set_partitions;
  std::vector<Basis>    Substitution;
  Type                  T1;
  Type                  T2;
  int                   Target_type_deg;  // from GenerateEquations.cc
  int                   Target_type_len;  // from GenerateEquations.cc
  struct polynomial*    The_ident;
  void*                 The_list;  // should be casted to Eqn_list_node<T>*

  // was static in ExtractMatrix.c; should be casted to Node<T>**
  void**                TheSparseMatrix;  

  // should be casted to Node<T>**
  void**                TheSparseMatrix_CreateMatrix;

  std::vector<Name>     Var_types;
  int                   Whatsleft;



  void  change_field (const FieldElemType, const bool);
  void  change_prime_field (const long long, const bool);
  void  change_number_field (const LiDIA::number_field&, const bool);
  void  AddSetPartition (const Name, const int);
  bool  AreBasisElements (const Degree) const;
  Basis BasisEnd (const Degree) const;
  Basis BasisStart (const Degree) const;
  bool  CreateBasisTable (int);
  bool  CreateColtoBP();
  void  DeleteSetPartition (const Name);
  int   DestroyBPtoCol();
  Basis EnterBasis (const Basis, const Basis, const Name);
  int   GetCol (const Basis, const Basis) const;
  Name  GetType (const Basis) const;
  int   GetVarNumber (const char) const;
  void  InitSeqSubtypes();
  void  InstallDegree1();
  Basis NextBasisSameType (const Basis) const;
  bool  OKSetPartitions() const;
  void  PrintBasis (const Basis, std::ostream&) const;
#if DEBUG_PERMUTATIONS
  void  PrintPermutation (const int Var_num, const Perm P) const;
  void  PrintPermutationList() const;
#endif
#if DEBUG_SEQ_SUBTYPES
  void  PrintSeqSubtypes() const;
#endif
#if DEBUG_SET_PARTITIONS
  void  PrintSetPartitions() const;
  void  PrintVarTypes() const;
#endif
#if DEBUG_SUBSTITUTION
  void  PrintSubstitution() const;
#endif
  void  Process (const Degree, const Degree, int*);
  void  UpdateDegToBasisTable (const int, const Basis);




  ////////////////////////////////////////////////////////////////////////////
  //                       AddAE                                            //
  ////////////////////////////////////////////////////////////////////////////
  // Add Alg_element *p1 to *p2
  //     *p2 = *p2 + *p1.                                         
  // We have to scan p2 to find and store p2->first.   
  // MODIFIES: *p2 -- Alg_element
  // RETURNS: 1 if successfull, 0 otherwise
  template<typename T> void AddAE (const typename Alg_element<T>::type& p1, 
								   typename Alg_element<T>::type& p2) const
  {
	typename std::map<Basis, T>::const_iterator p1i = p1.begin();
	typename std::map<Basis, T>::iterator p2i       = p2.begin();
    while (p1i != p1.end()) 
	  {
		if (p2i == p2.end()) 
		  {  
			// occurs if max_basis(p2) < max_basis(p1)
			p2[p1i->first] = p1i->second;
			p1i++;
		  } 
		else if (p2i->first == p1i->first) 
		  {
			p2i->second = (p2i->second) + (p1i->second);
			p1i++;
			p2i++;
		  } 
		else if (p2i->first < p1i->first) 
		  {
			p2i++;
		  } 
		else if (p2i->first > p1i->first) 
		  {
			p2[p1i->first] = p1i->second;
			p1i++;
		  }
	  }

    clear_zeros<T> (p2);
  } // end of Albert::AddAE()






  ////////////////////////////////////////////////////////////////////////////
  //                       AppendLocalListToTheList                         //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool AppendLocalListToTheList()
  {
	if ((this->Local_list == NULL) || (this->The_list == NULL))
	  {
		return (1);
	  }
	Basis_pair_node<T>* temp_ll = (Basis_pair_node<T>*)this->Local_list;    
	int ll_length = 1;
	while (temp_ll != NULL) 
	  {
		ll_length++;
		temp_ll = temp_ll->next;
	  }
	Eqn_list_node<T>* Temp_list = (Eqn_list_node<T>*)this->The_list;
	while (Temp_list->next != NULL)
	  {
		Temp_list = Temp_list->next;
	  }
	
	Temp_list->basis_pairs = 
	  (Basis_pair<T>*)(Mymalloc (ll_length * sizeof(Basis_pair<T>)));
	assert_not_null(Temp_list->basis_pairs);
	
	temp_ll = (Basis_pair_node<T>*)this->Local_list;
	int i;
	for (i=0; i < (ll_length-1); i++) 
	  {
		new (&(Temp_list->basis_pairs[i].coef)) T;  // "placement new"
		Temp_list->basis_pairs[i].coef = (temp_ll->bp).coef;    
		Temp_list->basis_pairs[i].left_basis  = (temp_ll->bp).left_basis;    
		Temp_list->basis_pairs[i].right_basis = (temp_ll->bp).right_basis;    
		temp_ll = temp_ll->next;
	  }
	new (&(Temp_list->basis_pairs[i].coef)) T;  // "placement new"
	Temp_list->basis_pairs[i].coef = (T)0;
	Temp_list->basis_pairs[i].left_basis = 0; 
	Temp_list->basis_pairs[i].right_basis = 0; 
	Temp_list->next = GetNewEqnListNode<T>();
	assert_not_null(Temp_list->next);
	return (1);
  } // end of Albert::AppendLocalListToTheList()





  ////////////////////////////////////////////////////////////////////////////
  //                       AppendToLocalList                                //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool AppendToLocalList (Basis_pair_node<T>* Rl)
  {
	assert_not_null(Rl);
	if (this->Local_list == NULL)
	  {
		this->Local_list = Rl; 
	  }
	else 
	  {
		Basis_pair_node<T>* temp_ll = (Basis_pair_node<T>*)this->Local_list;
		while (temp_ll->next != NULL)
		  {
			temp_ll = temp_ll->next;
		  }
		temp_ll->next = Rl; 
	  }
	return (1);
  } // end of Albert::AppendToLocalList()




  ////////////////////////////////////////////////////////////////////////////
  //                       CreateMultTable                                  //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> void CreateMultTable()
  {
	if (this->mult_table == NULL)
	  {
		this->mult_table = 
		  new std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >;
	  }
  } // end of Albert::CreateMultTable()






  ////////////////////////////////////////////////////////////////////////////
  //                       Delete_Element                                   //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> 
  void Delete_Element (Node<T>** Sparse_Matrix, int RowId, Node<T>* Prev_Ptr)
  {
	// Call primitive to delete element
	this->Delete_Node<T> (Sparse_Matrix, RowId, Prev_Ptr);
  }



  ////////////////////////////////////////////////////////////////////////////
  //                       Delete_Node                                      //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> 
  void Delete_Node (Node<T>** Sparse_Matrix, int RowId, Node<T>* Prev_Ptr)
  {
	Node<T>* Bad_Node;

	// It is the first element node in the row
	if (Prev_Ptr == NULL)
	  {
		Bad_Node = (Node<T>*) &(*Sparse_Matrix[RowId]);
		
		// it is not the only one in the row
		if (Bad_Node->Next_Node != NULL)
		  {
			Sparse_Matrix[RowId] = Bad_Node->Next_Node;
		  }
		// it is the only one in the row
		else
		  {
			Sparse_Matrix[RowId] = NULL;
		  }
	  }
	// There are other nodes before and after this one
	else
	  {
		Bad_Node = Prev_Ptr->Next_Node;
		Prev_Ptr->Next_Node = Bad_Node->Next_Node;
	  }
	
	// send the node back to the memory management system
	this->PutRecord<T> (Bad_Node);
  } // end of Albert::Delete_Node()





  ////////////////////////////////////////////////////////////////////////////
  //                       DestroyMultTable                                 //
  ////////////////////////////////////////////////////////////////////////////
  // Free the memory occupied by Multiplication table
  template<typename T> void DestroyMultTable()
  {
	if (this->mult_table != NULL)
	  {
		// not necessary?
		((std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*)(this->mult_table))->clear();

		delete (std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*)(this->mult_table);
	  }
  } // end of Albert::DestroyMultTable()





  ////////////////////////////////////////////////////////////////////////////
  //                       DestroySparseMatrix                              //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> void DestroySparseMatrix ()
  {
	mem_block<T>* block_ptr;
	mem_block<T>* prev_block_ptr;
	
	// if no blocks allocated then there is no matrix to destroy
	if (this->block_list_head == NULL)
	  {
		return;
	  }
	block_ptr      = (mem_block<T>*)this->block_list_head;
	prev_block_ptr = (mem_block<T>*)this->block_list_head;
	
	// walk the list deleting blocks behind us
	while (block_ptr != NULL)
	  {
		prev_block_ptr = block_ptr;
		block_ptr = block_ptr->next_block;
#ifdef DEBUGMM
		printf("Freeing block #%d\n",prev_block_ptr->id);
#endif
		free (prev_block_ptr);
	  }
	
	// free the row pointers
	free (this->Matrix_Base_Ptr);

	// null out static variables
	this->block_list_head = NULL;
	this->node_list_head = NULL;
	this->current_block = NULL;
#ifdef DEBUGMM
	blocknum = 0;
#endif
  } // end of Albert::DestroySparseMatrix()






  ////////////////////////////////////////////////////////////////////////////
  //                       DoCreateSubs                                     //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool DoCreateSubs (const int row, const int col)
  {
	if (row >= this->Num_vars) 
	  {
#if DEBUG_SUBSTITUTION
		this->PrintSubstitution();
#endif
		// Now for each substitution records, perform the substitution
		if (this->PerformSubs<T>() != 1)
		  {
			return (0);
		  }
	  }
	else if (col >= this->Deg_var[row])
	  {
		if (! this->DoCreateSubs<T> (row+1, 0))
		  {
			return (0);
		  }
	  }
	else 
	  {
		Basis b = 
		  FirstBasis (this->Set_partitions[col * (this->Num_vars) + row]); 
		bool rc;
		while (b != 0)
		  {
			this->Substitution[row * (this->Max_deg_var) + col] = b;
			rc = this->DoCreateSubs<T> (row, col+1);
			b = this->NextBasisSameType (b);
			if (! rc)
			  {
				return (0);
			  }
		  }
	  }
	return (1);
  } // end of Albert::DoCreateSubs()





  ////////////////////////////////////////////////////////////////////////////
  //                       DoPermutation                                    //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> void DoPermutation (const int row)
  {
	if (this->status_PerformSubs != 1)
	  {
		return;
	  }
	if (row == this->Num_vars) 
	  {
#if DEBUG_PERMUTATIONS
		this->PrintPermutationList();
#endif
		// Do the actual substitution using the current permutations of all
		// variables
		this->status_PerformSubs = this->Expand<T>();
	  }
	else 
	  {
		Perm pi = GetFirstPermutation (this->Deg_var[row]);
		if (pi == NULL)
		  {
			this->status_PerformSubs = 0;
		  }
		while ((pi != NULL) && (this->status_PerformSubs == 1)) 
		  {
			this->Permutation_list[row] = pi;
			this->DoPermutation<T> (row+1);
			pi = GetNextPermutation (pi, this->Deg_var[row]);
		  }
	  }
  } // end of Albert::DoPermutation()





  ////////////////////////////////////////////////////////////////////////////
  //                       EnterProduct                                     //
  ////////////////////////////////////////////////////////////////////////////
  // Enter the product of Basis B1 and B2 in the Multiplication table as 
  // Terms_list Tl.
  template<typename T> inline void EnterProduct
  (const Basis B1, const Basis B2, const std::vector<std::pair<T, Basis> >& Tl)
  {
	const std::pair<Basis, Basis> bb = std::make_pair (B1, B2);
	if (
		((std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*)(this->mult_table))->find (bb) 
		== 
		((std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*)(this->mult_table))->end()
		)
	  {
		// TODO:
		std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >* p = 
		  (std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*)(this->mult_table);
		(*p)[bb] = Tl;
	  } 
	else 
	  {
		// should never occur?
		std::cerr << "already present" << std::endl;
	  }
  } // end of Albert::EnterProduct()





  ////////////////////////////////////////////////////////////////////////////
  //                       Expand                                           //
  ////////////////////////////////////////////////////////////////////////////
  // We are getting to the core of the internals
  template<typename T> bool Expand()
  {
	if (this->The_ident == NULL)
	  {
		return (1);
	  }
	term_head* temp_head = this->The_ident->terms;
	Basis_pair_node<T>* temp_list;
	while (temp_head != NULL) 
	  {
		this->running_list = NULL;
		if (this->SubstituteWord<T> (temp_head->term) != 1)
		  {
			return (0);
		  }
		temp_list = (Basis_pair_node<T>*)this->running_list; 
		while (temp_list != NULL) 
		  {
			(temp_list->bp).coef *= (T)(temp_head->coef);
			temp_list = temp_list->next;
		  }
		this->AppendToLocalList<T> ((Basis_pair_node<T>*)this->running_list);
		temp_head = temp_head->next;
	  }
	return (1);
  } // end of Albert::Expand()





  ////////////////////////////////////////////////////////////////////////////
  //                       ExpandTerm                                       //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> int ExpandTerm (typename Alg_element<T>::type& Ans, 
									   term_node* W, bool* status) const
  {
	if (*status != 1)
	  {
		return (-1);
	  }
	if ((W->left == NULL) && (W->right == NULL)) 
	  {
		const Basis b = GetBasisNumberofLetter (W->letter);
		SetAE<T> (Ans, b, (T)1);
	  }
	else 
	  {
		typename Alg_element<T>::type left;
		typename Alg_element<T>::type right;
		if (*status == 1)
		  {
			this->ExpandTerm<T> (left, W->left, status);
		  }
		if (*status == 1)
		  {
			this->ExpandTerm<T> (right, W->right, status);
		  }
		*status = this->MultAE<T> (left, right, Ans);
	  }
	return (0);
  } // end of Albert::ExpandTerm()





  ////////////////////////////////////////////////////////////////////////////
  //                       FillPairPresent                                  //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool FillPairPresent()
  {
	Eqn_list_node<T>* temp = (Eqn_list_node<T>*)this->First_eqn_list_node;
	int i;
	while (temp != NULL) 
	  {
		i=0;
		assert_not_null(temp->basis_pairs);
		while (temp->basis_pairs[i].coef != (T)0) 
		  {
			this->Pair_present.insert 
			  (std::make_pair (temp->basis_pairs[i].left_basis,
							   temp->basis_pairs[i].right_basis));
			i++;
		  }
		this->Num_equations++;
		temp = temp->next;
	  }
	return (1);
  }





  ////////////////////////////////////////////////////////////////////////////
  //                       Gen                                              //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> void Gen (const Name n, const int d, const int j)
  {
	if (this->status_Multpart != 1)
	  {
		return;
	  }
	if (d == 1) 
	  {
		this->AddSetPartition (n, j);
		this->SplitJthType<T> (j+1);
		this->DeleteSetPartition (j);
	  }
	else 
	  {
		const int degn = GetDegreeName (n);
		const int lower = (((degn % d) == 0) ? degn/d : degn/d + 1);
		const int upper = degn - d + 1;
		Name n1, n_minus_n1;
		for (int i=lower; i <= upper; i++) 
		  {
			n1 = FirstTypeDegree (i);
			while ((n1 != -1) && (this->status_Multpart == 1)) 
			  {
				if (IsSubtype (n1, n)) 
				  {
					this->AddSetPartition (n1, j);
					SubtractTypeName (n, n1, &n_minus_n1);
					this->Gen<T> (n_minus_n1, d-1, j);
					this->DeleteSetPartition (j);
				  }
				n1 = NextTypeSameDegree (n1);
			  }
		  }
	  }
  } // end of Albert::Gen()






  ////////////////////////////////////////////////////////////////////////////
  //                       GenerateEquations                                //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool GenerateEquations (const Name N)
  {
	if (this->The_list == NULL)
	  {
		return (1);
	  }
	this->Target_type_len = GetTargetLen(); 
	this->Target_type_deg = GetDegreeName (N);
	this->Num_vars = 0;
	//int i,j;
	int i;
	for (i=0; i<NUM_LETTERS; i++)
	  {
		if (this->The_ident->deg_letter[i] > 0)
		  {
			this->Num_vars++;
		  }
	  }
	
	GetNewType (this->Target_type);
	this->Seq_sub_types.resize ((this->Num_vars) * (this->Target_type_len));
	this->Deg_var.clear();
	this->Cur_deg_vars.clear();

	NameToType (N, this->Target_type);
	
	//for (i=0,j=0; i<NUM_LETTERS; i++)
	for (i=0; i<NUM_LETTERS; i++)
	  {
		if (this->The_ident->deg_letter[i] > 0)
		  {
			this->Deg_var.push_back (this->The_ident->deg_letter[i]);
			//j++;
		  }
	  }
	
	this->InitSeqSubtypes();
	
	// Starting of deep recursive calls
	this->GenerateSeqSubtypes<T> (0, 0, 0);  
	
	return (this->status_GenerateEquations);
  } // end of Albert::GenerateEquations()






  ////////////////////////////////////////////////////////////////////////////
  //                       GenerateSeqSubtypes                              //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> void GenerateSeqSubtypes (const int Cur_row, 
												 const int Cur_col, 
												 const int Weight)
  {
	if (this->status_GenerateEquations != 1)
	  {
		return;
	  }
	int whatsave, tsave, csave;
#if DEBUG_SEQ_SUBTYPES
	static int count = 1;
#endif
	
	if (Cur_col == this->Target_type_len) 
	  {
#if DEBUG_SEQ_SUBTYPES
		std::cout << "Printing " << count++ << "th SeqSubtypes " << std::endl;
		this->PrintSeqSubtypes();
#endif
		// Now we are jumping into another module while inside a recursive call
		this->status_GenerateEquations = this->PerformMultiplePartition<T> ();
	  }
	else if (Cur_row == (this->Num_vars - 1)) 
	  {
		if ((this->Whatsleft + this->Cur_deg_vars[Cur_row]) >= 
			this->Deg_var[Cur_row]) 
		  {
			csave = this->Cur_deg_vars[Cur_row];
			whatsave = this->Whatsleft;
			tsave = 
			  this->Seq_sub_types[Cur_row * (this->Target_type_len) + Cur_col]; 
			this->Seq_sub_types[Cur_row * (this->Target_type_len) + Cur_col] = 
			  this->Target_type[Cur_col] - Weight;
			this->Cur_deg_vars[Cur_row] += 
			  this->Seq_sub_types[Cur_row * (this->Target_type_len) + Cur_col];
			this->Whatsleft -= 
			  this->Seq_sub_types[Cur_row * (this->Target_type_len) + Cur_col];
			if ((Cur_col < (this->Target_type_len - 1)) ||
				((Cur_col == (this->Target_type_len - 1)) && 
				 (this->Cur_deg_vars[Cur_row] >= this->Deg_var[Cur_row])))
			  {
				this->GenerateSeqSubtypes<T> (0, Cur_col+1, 0);
			  }
			this->Seq_sub_types[Cur_row * (this->Target_type_len) + Cur_col] = 
			  tsave; 
			this->Cur_deg_vars[Cur_row] = csave; 
			this->Whatsleft = whatsave;
		  }
	  }
	else 
	  {
		for (int i = this->Target_type[Cur_col] - Weight; i>=0; i--) 
		  {
			if ((this->Cur_deg_vars[Cur_row] + this->Whatsleft) >= 
				this->Deg_var[Cur_row]) 
			  {
				csave = this->Cur_deg_vars[Cur_row];
				whatsave = this->Whatsleft;
				tsave = this->Seq_sub_types
				  [Cur_row * (this->Target_type_len) + Cur_col]; 
				this->Seq_sub_types[Cur_row * (this->Target_type_len) + Cur_col]
				  = i; 
				this->Cur_deg_vars[Cur_row] += i; 
				this->Whatsleft -= i; 
                
				if ((Cur_col < (this->Target_type_len - 1)) ||
					((Cur_col == (this->Target_type_len - 1)) && 
					 (this->Cur_deg_vars[Cur_row] >= this->Deg_var[Cur_row])))
				  {
					this->GenerateSeqSubtypes<T> (Cur_row+1, Cur_col, Weight+i);
				  }
				this->Seq_sub_types[Cur_row * (this->Target_type_len) + Cur_col]
				  = tsave; 
				this->Cur_deg_vars[Cur_row] = csave; 
				this->Whatsleft = whatsave;
			  }
		  }
	  }
  } // end of Albert::GenerateSeqSubtypes()






  ////////////////////////////////////////////////////////////////////////////
  //                       GetRecord                                        //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> Node<T>* GetRecord()
  {
	// if there are no free nodes available then allocate new ones
	if (this->node_list_head == NULL)
	  {
		mem_block<T>* new_block = 
		  (mem_block<T>*)Mymalloc (sizeof(mem_block<T>));
#ifdef DEBUGMM
		new_block->id = blocknum;
		printf ("Allocated block #%d\n", new_block->id);
		blocknum++;
#endif
		new_block->next_block = NULL;
		
		// if this is the first block make it the head of the list
		if (this->block_list_head == NULL)
		  {
			this->block_list_head = new_block;
			this->current_block   = new_block;
		  }
		// else connect it to the list and make it current
		else
		  {
			((mem_block<T>*)this->current_block)->next_block = new_block;
			this->current_block = 
			  ((mem_block<T>*)this->current_block)->next_block;
			this->current_block = new_block; 
		  }
		this->node_list_head = new_block->record;
		
		// carve it up into a linked list
		int i;
		for (i=0; i < (RECORDS_PER_BLOCK-1); i++)
		  { 
			new_block->record[i].Next_Node = &new_block->record[i+1];
		  }
		new_block->record[i].Next_Node = NULL;
	  }
	
	// return the front node in the list
	Node<T>* return_value = (Node<T>*)this->node_list_head;
	
	// move the pointer to the next one
	this->node_list_head = ((Node<T>*)this->node_list_head)->Next_Node;
	
	return (return_value);
  } // end of Albert::GetRecord() 






  ////////////////////////////////////////////////////////////////////////////
  //                       Insert_Element                                   //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> inline
  void Insert_Element (Node<T>** Sparse_Matrix, const int matrow, 
					   const T& element, const int column, Node<T>* Prev_Ptr)
  {
	// get a new node structure from the memory management system
	Node<T>* New_Node = (Node<T>*)this->GetRecord<T>();
	
	// set the values of the structure
	new (&(New_Node->element)) T;   // "placement new"
	New_Node->element = element;
	New_Node->column  = column;
	
	// insert it into the correct spot in the matrix
	Insert_Node<T> (Sparse_Matrix, matrow, New_Node, Prev_Ptr);
  }



  
  ////////////////////////////////////////////////////////////////////////////
  //                       LeftTapAE                                        //
  ////////////////////////////////////////////////////////////////////////////
  // Multiply term with Alg_element *p1 and add to *p2.
  //         *p2 = xb.(*p1) + *p2            
  // RETURNS: 1 if successfull, 0 otherwise
  template<typename T> bool LeftTapAE (const T& x, const Basis b, 
									   const typename Alg_element<T>::type& p1, 
									   typename Alg_element<T>::type& p2) const
  {
	if ((x == (T)0) || IsZeroAE<T> (p1))
	  {
		return (1);
	  }
	else 
	  {
		bool status = 1;
		typename std::map<Basis, T>::const_iterator i;
		for (i = p1.begin(); i != p1.end(); i++) 
		  {
			if (i->second != (T)0)
			  {
				if (status == 1)
				  {
					status = this->Mult2basis<T> 
					  (b, i->first, x * (i->second), p2);
				  }
			  }
		  }
		clear_zeros<T> (p2);
		return (status);
	  }
  } // end of Albert::LeftTapAE()





  ////////////////////////////////////////////////////////////////////////////
  //                       Locate_Element                                   //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> Node<T>* Locate_Element (Node<T>* Tmp_Ptr, 
												const unsigned int row, 
												const unsigned int col) const
  {
	Node<T>* Prev_Ptr = NULL;
	Node<T>* Curr_Ptr;
	if (Tmp_Ptr == NULL)
	  {
		Curr_Ptr = ((Node<T>**)this->Matrix_Base_Ptr)[row];
	  }
	else
	  {
		Curr_Ptr = Tmp_Ptr;
	  }
	
	// We will either be at the end of the linked list or at the node just 
	// before the row,col that we want
	while ((Curr_Ptr != NULL) && (col > Curr_Ptr->column))
	  {
		Prev_Ptr = Curr_Ptr;
		Curr_Ptr = Curr_Ptr->Next_Node;
	  }
	return (Prev_Ptr);
  }




  ////////////////////////////////////////////////////////////////////////////
  //                       Mult2basis                                       //
  ////////////////////////////////////////////////////////////////////////////
  // RETURNS: Pointer to Terms_list, product of B1 and B2
  template<typename T> bool Mult2basis (const Basis B1, const Basis B2, 
										const T& x, 
										typename Alg_element<T>::type& P) const
  {
	const std::pair<Basis, Basis> bb = std::make_pair (B1, B2);
	typename std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >::const_iterator ii = 
	  ((std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*)(this->mult_table))->find (bb);

	if (
		ii == 
		((std::map<std::pair<Basis, Basis>, std::vector<std::pair<T, Basis> > >*)(this->mult_table))->end()
		)
	  {
		return (0);
	  }
	typename std::vector<std::pair<T, Basis> >::const_iterator jj;
	for (jj = ii->second.begin(); jj != ii->second.end(); jj++) 
	  {
		SetAE<T> (P, jj->second, GetAE<T> (P, jj->second) + x * (jj->first));
	  }
	return (1);
  } // end of Albert::Mult2basis




  ////////////////////////////////////////////////////////////////////////////
  //                       MultAE                                          //
  ////////////////////////////////////////////////////////////////////////////
  // Multiply Alg_elements *p1,*p2 and add to *p3.
  //          *p3 = (*p1) * (*p2) + *p3                            
  // RETURNS: 1 if successfull, 0 otherwise
  template<typename T> 
  bool MultAE (const typename Alg_element<T>::type& p1, 
			   const typename Alg_element<T>::type& p2, 
			   typename Alg_element<T>::type& p3) const
  {
	if (IsZeroAE<T> (p1) || IsZeroAE<T> (p2))
	  {
		return (1);
	  }
	else 
	  {
		bool status = 1;
		typename std::map<Basis, T>::const_iterator p1i;
		for (p1i = p1.begin(); p1i != p1.end(); p1i++) 
		  {
			if (p1i->second != (T)0)
			  {
				if (status)
				  {
					status = this->LeftTapAE (p1i->second, p1i->first, p2, p3);
				  }
			  }
		  }
		clear_zeros<T> (p3);
		return (status);
	  }
  } // end of Albert::MultAE()





  ////////////////////////////////////////////////////////////////////////////
  //                       ProcessIndependentBasis                          //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool ProcessIndependentBasis()
  {
	if (this->Num_cols_ExtractMatrix == 0)
	  {
		return (0);
	  }
	std::vector<std::pair<T, Basis> > tl(1);
	Basis n, b1, b2;
	for (unsigned int j=0; j < this->Num_cols_ExtractMatrix; j++) 
	  {
		if (! this->Dependent[j]) 
		  {
			b1 = this->ColtoBP[j].left_basis;
			b2 = this->ColtoBP[j].right_basis;
			n = EnterBasis (b1, b2, TypeToName (this->Cur_type));

			tl[0].first  = (T)1;
			tl[0].second = n;
			this->EnterProduct<T> (b1, b2, tl);

			this->BasisNames[j] = n;
		  }
	  }
	return (1);
  } // end of ProcessIndependentBasis()





  ////////////////////////////////////////////////////////////////////////////
  //                       PerformMultiplePartition                         //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool PerformMultiplePartition ()
  {
	this->status_Multpart = 1;
	this->Var_types.clear();

	// these two were static in Multpart.cc; used only in this func
	Type temp_type_Multpart (NUM_LETTERS, 0);
	const int target_type_len = GetTargetLen();

	int i, j;
	for (i=0; i<target_type_len; i++)
	  {
		temp_type_Multpart[i] = 0;
	  }
	
	for (i=0; i < this->Num_vars; i++) 
	  {
		for (j=0; j < target_type_len; j++)
		  {
			temp_type_Multpart[j] = this->Seq_sub_types[i*target_type_len + j];
		  }
		this->Var_types.push_back (TypeToName (temp_type_Multpart));
	  }
	this->Max_deg_var = this->Deg_var[0];
	for (i=1; i < this->Num_vars; i++)
	  {
		if (this->Deg_var[i] > this->Max_deg_var)
		  {
			this->Max_deg_var = this->Deg_var[i];
		  }
	  }

	this->Set_partitions.resize ((this->Max_deg_var) * (this->Num_vars));
	this->Cur_index_var.clear();
	for (i=0; i < this->Max_deg_var; i++)
	  {
		for (j=0; j < this->Num_vars; j++)
		  {
			this->Set_partitions[i * this->Num_vars + j] = 0;
		  }
	  }
	for (i=0; i < this->Num_vars; i++)
	  {
		this->Cur_index_var.push_back (0);
	  }
	
#if DEBUG_SET_PARTITIONS
	PrintVarTypes();
#endif
	this->SplitJthType<T> (0);    // Start a recursive call
	
	return (this->status_Multpart);    
  } // end of Albert::PerformMultiplePartition()






  ////////////////////////////////////////////////////////////////////////////
  //                       PerformSubs                                      //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool PerformSubs ()
  {
	this->status_PerformSubs = 1;
	this->Permutation_list.resize (this->Num_vars);

	this->Local_list = NULL;
	this->DoPermutation<T> (0);    // Start of recursive call
	if (this->status_PerformSubs != 1)
	  {
		return (0);
	  }
	if (this->AppendLocalListToTheList<T>() != 1)
	  {
		return (0);
	  }
	FreeLocalList<T> ((Basis_pair_node<T>*)this->Local_list);

	//FreePermutationList (this->Permutation_list);
	if (! this->Permutation_list.empty())
	  {
		for (int i=0; i < this->Num_vars; i++)
		  {
			free (this->Permutation_list[i]);
		  }
	  }

	//free (this->Permutation_list);
	this->Permutation_list.clear();
	return (1);
  } // end of Albert::PerformSubs()





  ////////////////////////////////////////////////////////////////////////////
  //                       Print_MultTable                                  //
  ////////////////////////////////////////////////////////////////////////////
  // arg: 
  // output_for_gap - whether to format output for GAP (1) or for human (0)
  template<typename T> 
  void Print_MultTable (std::ostream& os, const bool output_for_gap) const
  {
	typename Alg_element<T>::type prod;

	unsigned int dim = this->Basis_table.size() - 1;
	if (dim > 0)
	  {
		if (output_for_gap)
		  {
			if (this->field_elem_type == LIDIA_TYPE)
			  {
				os << "x := Indeterminate (Rationals, \"x\");" << std::endl
				   << "p := " << Albert::NF << ";" << std::endl
				   << "F := AlgebraicExtension (Rationals, p);" << std::endl;
			  }
			os << "T := EmptySCTable (" << dim << ", Zero(";
			this->print_field (os, 1);
			os << "));" << std::endl;
		  }
		else
		  {
			os << std::endl << "Multiplication table: " << std::endl;
		  }
		unsigned int j;
		for (unsigned int i=1; i <= dim; i++) 
		  {
			for (j=1; j <= dim; j++) 
			  {
				prod.clear();
				this->Mult2basis<T> (i, j, (T)1, prod);
				if (! IsZeroAE<T> (prod))
				  {
					if (output_for_gap)
					  {
						os << "SetEntrySCTable (T, " << i << ", " << j << ", [";
					  }
					else
					  {
						os << "(b" << i << ")*(b" << j << ")" << std::endl;
					  }
					// printing of element of lidia type is not working yet
					Print_AE<T> (prod, os, output_for_gap);
					if (output_for_gap)
					  {
						os << "]);" << std::endl;
					  }
					else
					  {
                        os << std::endl << std::endl;
					  }
				  }
			  }
		  } // end of double loop thru basis elements
		if (output_for_gap)
		  {
			os << "A := AlgebraByStructureConstants (";
			this->print_field (os, 1);
			os << ", T); " << std::endl;
		  }
	  }
  } // end of Albert::Print_MultTable()





  ////////////////////////////////////////////////////////////////////////////
  //                       PrintEqns                                        //
  ////////////////////////////////////////////////////////////////////////////
#if DEBUG_EQNS
  template<typename T> void PrintEqns ()
  {
	assert_not_null_retvoid(this->The_list);
	Eqn_list_node<T>* temp = (Eqn_list_node<T>*)this->The_list;
	int count = 1;
	std::cout << std::endl << " The Equations are : " << std::endl;
	while (temp != NULL) 
	  {
		//printf("%2d.  ",count++);
		std::cout << count++ << ".  ";
		PrintEqn (temp->basis_pairs);
		temp = temp->next;
	  }
  }
#endif




  ////////////////////////////////////////////////////////////////////////////
  //                       ProcessDegree                                    //
  ////////////////////////////////////////////////////////////////////////////
  // Process all Types of degree i
  template<typename T> bool ProcessDegree (const Degree i)
  {
	Name n;
	bool status = 1;
	Basis begin_basis;
	Basis end_basis = 0;
	if (i == 1)
	  {
		this->InstallDegree1();
	  }
	else 
	  {
		n = FirstTypeDegree (i);
		while ((status == 1) && (n != -1)) 
		  {
			begin_basis = this->Basis_table.size();
			status = this->ProcessType<T> (n);
			end_basis = this->Basis_table.size() - 1;
			if (end_basis < begin_basis)
			  {
				UpdateTypeTable (n, 0, 0);  // No Basis table entries
			  }
			else
			  {
				UpdateTypeTable (n, begin_basis, end_basis);
			  }
			n = NextTypeSameDegree (n);
		  }
		this->Current_dimension = end_basis;
	  }
	return (status);
  } // end of Albert::ProcessDegree()




  ////////////////////////////////////////////////////////////////////////////
  //                       ProcessOtherIndependentBasis                     //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> void ProcessOtherIndependentBasis (const unsigned int J)
  {

	std::vector<std::pair<T, Basis> > tl(1);

	if (this->Cur_type.size() == J)
	  {
		const int deg = GetDegree (this->T1);
		if ((deg > 0) && (deg < this->Cur_type_degree)) 
		  {
			for (unsigned int i=0; i < this->Cur_type.size(); i++)
			  {
				this->T2[i] = this->Cur_type[i] - this->T1[i];
			  }
			Basis m1 = BeginBasis (TypeToName (this->T1));
			Basis m2 = EndBasis   (TypeToName (this->T1));
			Basis n1 = BeginBasis (TypeToName (this->T2));
			Basis n2 = EndBasis   (TypeToName (this->T2));
			Basis n;
			if ((0 < m1) && (m1 <= m2) && (0 < n1) && (n1 <= n2)) 
			  {
				Basis j;
				for (Basis i = m1; i <= m2; i++) 
				  {
					for (j = n1; j <= n2; j++) 
					  {
						if (this->GetCol (i, j) == -1) 
						  {
							n = EnterBasis (i, j, TypeToName (this->Cur_type));
							
							tl[0].first  = (T)1;
							tl[0].second = n;
							this->EnterProduct<T> (i, j, tl);
						  }
					  }
				  }
			  }
		  }
	  }
	else 
	  {
		int save;
		for (Degree i=0; i <= this->Cur_type[J]; i++) 
		  {
			save = this->T1[i];
			this->T1[J] = i;
			this->ProcessOtherIndependentBasis<T> (J+1);
			this->T1[i] = save;
		  }
	  }
  } // end of Albert::ProcessOtherIndependentBasis()





  ////////////////////////////////////////////////////////////////////////////
  //                       ProcessType                                      //
  ////////////////////////////////////////////////////////////////////////////
  // For each identity f, whose degree is less than the degree of type t, 
  // generate equations corresponding to f.
  // Then solve those equations, to get New Basis and write other basis pairs 
  // in terms of existing basis.
  template<typename T> bool ProcessType (const Name n)
  {
	// Header record of linked list
	this->The_list = GetNewEqnListNode<T>();
	assert_not_null(this->The_list);

	struct id_queue_node* temp_id_node = this->Id_queue.first;
	bool status = 1;
	while ((temp_id_node != NULL) && (status == 1)) 
	  {
		this->The_ident = temp_id_node->identity;
		if ((status == 1) && (this->The_ident->degree <= GetDegreeName (n)))
		  {
			status = this->GenerateEquations<T> (n);
		  }
		temp_id_node = temp_id_node->next;
	  }
	
#if DEBUG_EQNS
	PrintEqns<T>();
#endif
	
	if (status == 1)
	  {
		status = this->SolveEquations<T> (n);
	  }
	if (this->The_list != NULL)
	  {
		FreeEqns<T> ((Eqn_list_node<T>*)(this->The_list));
	  }
	return (status);
  } // end of Albert::ProcessType()






  ////////////////////////////////////////////////////////////////////////////
  //                       PutRecord                                        //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> void PutRecord (Node<T>* freed_node)
  {
	// Null out information in the node to be safe
	freed_node->element = (T)0;
	freed_node->column = 0;     
	freed_node->Next_Node = NULL;
	
	// return it to the front of the list
	freed_node->Next_Node = (Node<T>*)this->node_list_head;
	this->node_list_head = freed_node;
  }




  ////////////////////////////////////////////////////////////////////////////
  //                       SolveEquations                                   //
  ////////////////////////////////////////////////////////////////////////////
  // Convert the given list of equations into Matrix, i.e. one row for each 
  // equation and one column for each unique basis pair present in all 
  // equations.
  // Then Reduce that Matrix into row canonical form.
  // Then Extract from the Reduced Matrix i.e. Find New Basis and enter them 
  // into Basis Table. 
  // Then write Dependent Basis pairs into Basis by entering products into 
  // Mult_table.
  //
  // Called once.
  //
  // arg: t -- Current Type being processed
  template<typename T> bool SolveEquations (const Name n)
  {
	// Size of matrix
	unsigned int rows = 0;
	unsigned int cols = 0;
	unsigned int rank = 0;
	this->Matrix_Base_Ptr = NULL;

	Unique_basis_pair_list BPCptr = NULL;  // pointer to BPtoCol
	bool status = 1;

	// this flag is only set on the last degree of the generator
	if (this->gather_density_flag)
	  {
		this->num_elements     = 0;
		this->max_num_elements = 0;
		this->matrix_size      = 0;
	  }
	// CreateMatrix will initialize rows, cols, m, BP 
	// and fill in matrix and BPtoCol
	status = this->SparseCreateTheMatrix<T> (&rows, &cols, &BPCptr, n);
	
	// make sure that we get the correct information from creating the 
	// matrix for our statistics
	if (this->gather_density_flag)
	  {
		this->max_num_elements = this->num_elements;
		this->matrix_size = rows * cols;
	  }
	
#if DEBUG_MATRIX
	PrintColtoBP();
	PrintTheMatrix();
#endif
	
	if (status == 1)
	  {
		status = this->SparseReduceMatrix<T> (rows, cols, &rank);
	  }
	
#if DEBUG_MATRIX
	PrintTheRMatrix();
#endif
	
	// ExtractMatrix will expand basis table & MultTable !
	if (status == 1) 
	  {
		this->Num_rows_ExtractMatrix = rows;
		this->Num_cols_ExtractMatrix = cols;
		this->MatrixRank = rank;
		this->ColtoBP = BPCptr;
		status = this->SparseExtractFromMatrix<T> (n);
	  }
#if DEBUG_MATRIX
	PrintDependent();
#endif
	
	DestroyBPtoCol();    // Make sure not null before malloc
	if (status == 1) 
	  {
		this->DestroySparseMatrix<T>();
	  }
	return (status);
  } // end of Albert::SolveEquations()






  ////////////////////////////////////////////////////////////////////////////
  //                       SparseAddRow                                     //
  ////////////////////////////////////////////////////////////////////////////
  /* Three things can happen with the SparseAddRow routine ....
	 First there is a target row and a row which is multiplied by a factor
	 and added to the target row. 
	 1. The result is nonzero and there is no column in the target row so add 
	 a new node.
	 2. The result is nonzero and there is a column in the target row so change
	 the value in the node.
	 3. The result is zero and there is a column in the target row so delete
	 the node.
  */
  template<typename T> void SparseAddRow (const T& Factor, 
										  const unsigned int Row1, 
										  const unsigned int Row2)
  {
	if (Factor == (T)0)
	  {
		return;
	  }
	
	// get the beginning of the two rows to work with
	Node<T>* Row1_Ptr = ((Node<T>**)this->Matrix_Base_Ptr)[Row1];
	Node<T>* Temp_Ptr = ((Node<T>**)this->Matrix_Base_Ptr)[Row2];
	
	// Process the linked list representing the row to be multiplied and 
	// to be added
	T TmpRow1_S_Product, TmpRow2_S_Sum;
	Node<T>* Prev_Ptr;
	Node<T>* Look_Ahead_Ptr;
	while (Row1_Ptr != NULL)
	  {
		// go ahead and perform the scalar multiplication
		TmpRow1_S_Product = Factor * (Row1_Ptr->element);
		
		// Try to get the previous node to the one that we really want which is
		// the one that would match the column value of the one that we wish to
		// add
		Prev_Ptr = this->Locate_Element<T> (Temp_Ptr, Row2, Row1_Ptr->column);
		Temp_Ptr = Prev_Ptr;
		
		// if there is no previous and the row is not empty then we want to work
		// on the first node in the row
		if ((Prev_Ptr == NULL) && 
			(! Row_empty<T> ((const Node<T>**)this->Matrix_Base_Ptr, Row2)))
		  {
			Look_Ahead_Ptr = ((Node<T>**)this->Matrix_Base_Ptr)[Row2];
		  }
		else
		  {
			if (Row_empty<T> ((const Node<T>**)this->Matrix_Base_Ptr, Row2))
			  {
				Look_Ahead_Ptr = NULL;
			  }
			else
			  {
				// Row is not empty so we must be before the node we want to
				// work with so if there is a next node that is the one we want
				if (Prev_Ptr->Next_Node != NULL)
				  {
					Look_Ahead_Ptr = Prev_Ptr->Next_Node;
				  }
				// there is not another one in the row so we want to work with 
				// this one
				else
				  {
					Look_Ahead_Ptr = Prev_Ptr;
				  }
			  }
		  }
		
		if (Look_Ahead_Ptr != NULL)
		  {
			// Row wasn't empty
			if (Look_Ahead_Ptr->column != Row1_Ptr->column)
			  {
				// The row doesn't have a node representing the actual column 
				// that we want so lets add one
				TmpRow2_S_Sum = TmpRow1_S_Product;
				this->Insert_Element<T> ((Node<T>**)this->Matrix_Base_Ptr, Row2,
										 TmpRow2_S_Sum, Row1_Ptr->column, 
										 Prev_Ptr);
				
				// gather statistics
				if (this->gather_density_flag)
				  {
					this->num_elements++;
					if (this->num_elements > this->max_num_elements)
					  {
						this->max_num_elements = this->num_elements;
					  }
				  }
			  }
			else
			  {
				// the column we want is there 
				TmpRow2_S_Sum = 
				  (Look_Ahead_Ptr->element) + TmpRow1_S_Product;
				if (TmpRow2_S_Sum == (T)0)
				  {
					// the result is zero so we must delete it
					this->Delete_Element<T> 
					  ((Node<T>**)this->Matrix_Base_Ptr, Row2, Prev_Ptr);
					// statistics
					if (this->gather_density_flag)
					  {
						this->num_elements--;
					  }
				  }
				else
				  {
					// the result is not zero and column is present so change 
					// value
					Change_Element<T> (Look_Ahead_Ptr, TmpRow2_S_Sum);
				  }
			  }
		  }
		else
		  {
			// We know the row is empty and we will add a node if the result is 
			// nonzero
			TmpRow2_S_Sum = TmpRow1_S_Product;
			
			// the result is nonzero so insert node
			this->Insert_Element<T> ((Node<T>**)this->Matrix_Base_Ptr, Row2, 
									 TmpRow2_S_Sum, Row1_Ptr->column, Prev_Ptr);
			
			// statistics 
			if (this->gather_density_flag)
			  {
				this->num_elements++;
				if (this->num_elements > this->max_num_elements)
				  {
					this->max_num_elements = this->num_elements;
				  }
			  }
		  }
		Row1_Ptr = Row1_Ptr->Next_Node;
	  }
  } // end of Albert::SparseAddRow()






  ////////////////////////////////////////////////////////////////////////////
  //                       SparseCreateTheMatrix                            //
  ////////////////////////////////////////////////////////////////////////////
  // This is virtually identical to CreateTheMatrix() except that we call 
  // SparseFillTheMatrix() and copy the Matrix ptr to a pointer internal to this
  // module to be altered and then copy it back before returning to 
  // SolveEquations()
  template<typename T> bool SparseCreateTheMatrix 
  (unsigned int* Rows, unsigned int* Cols, Unique_basis_pair_list* BPCptr, 
   const Name n)
  {
	this->Num_equations = 0;
	this->ColtoBP_CreateMatrix = NULL;
	this->TheSparseMatrix_CreateMatrix = NULL;
	this->N_CreateMatrix = n;
	this->First_eqn_list_node = this->The_list;
	
	if (this->The_list == NULL)
	  {
		return (1);
	  }
	if (((Eqn_list_node<T>*)(this->The_list))->basis_pairs == NULL)
	  {
		return (1);
	  }
	
	this->Pair_present.clear();
	this->FillPairPresent<T>();
	// PrintPairPresent();
	if (this->CreateColtoBP() != 1)
	  {
		return (0);
	  }
	if (this->SparseFillTheMatrix<T>() != 1)
	  {
		return(0);
	  }

	// pass locally derived values back to the calling routine
	this->Matrix_Base_Ptr = this->TheSparseMatrix_CreateMatrix;
	*BPCptr = this->ColtoBP_CreateMatrix;
	*Rows   = this->Num_equations;
	*Cols   = this->Pair_present.size(); 
	return (1);
  } // end of Albert::SparseCreateTheMatrix()






  ////////////////////////////////////////////////////////////////////////////
  //                       SparseExtractFromMatrix                          //
  ////////////////////////////////////////////////////////////////////////////
  // This is virtually identical to ExtractFromMatrix() except for the calls to
  // SparseFillDependent() and SparseProcessDependentBasis(). Operation are 
  // performed on a locally visible pointer to the matrix and then that pointer
  // is copied to the one passed in upon exit 
  template<typename T> bool SparseExtractFromMatrix (const Name N)
  {
	this->TheSparseMatrix = this->Matrix_Base_Ptr;
	//this->Dependent = NULL;
	//this->BasisNames = NULL;
	this->Dependent.clear();
	this->BasisNames.clear();
	const unsigned int len = GetTargetLen();
	this->Cur_type.resize (len);
	this->T1.resize (len);
	NameToType (N, this->T1);
	NameToType (N, this->Cur_type);
	this->Cur_type_degree = GetDegree (this->Cur_type);
	this->T2.resize (len);
	
	if (this->Num_cols_ExtractMatrix > 0 ) 
	  {
		this->Dependent.clear();
		unsigned int i;
		for (i=0; i < this->Num_cols_ExtractMatrix; i++)
		  {
			this->Dependent.push_back (0);
		  }
		this->BasisNames.clear();
		for (i=0; i < this->Num_cols_ExtractMatrix; i++)
		  {
			this->BasisNames.push_back (0);
		  }
		this->SparseFillDependent<T>();
		this->ProcessIndependentBasis<T>();
		this->SparseProcessDependentBasis<T>();

		//if (! this->Dependent.empty())
		//{
			this->Dependent.clear();
			//}

		this->BasisNames.clear();
	  }
	this->ProcessOtherIndependentBasis<T> (0);
	return (1);
  } // end of Albert::SparseExtractFromMatrix()





  ////////////////////////////////////////////////////////////////////////////
  //                       SparseFillDependent                              //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool SparseFillDependent()
  {
	if ((this->Num_rows_ExtractMatrix == 0) || 
		(this->Num_cols_ExtractMatrix == 0))
	  {
		return (0);
	  }
	// This routine is much simpler than its sister FillDependent() since the 
	// first node in the linked list of each row will contain the first nonzero 
	// element.
	Node<T>* Tmp_Ptr;
	unsigned int i=0;
	while (i < this->MatrixRank)
	  {
		Tmp_Ptr = ((Node<T>**)this->TheSparseMatrix)[i];
		
		// place ones in the correct columns of the Dependent structure
		this->Dependent[Tmp_Ptr->column] = 1;
		i++;
	  } 
	return (1);
  } // end of SparseFillDependent()




  ////////////////////////////////////////////////////////////////////////////
  //                       SparseFillTheMatrix                              //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool SparseFillTheMatrix()
  {
	if ((this->Pair_present.size() == 0) || (this->Num_equations == 0))
	  {
		return (1);
	  }

	T Tmp_S_Sum;
	Node<T>* Prev_Ptr;
	Node<T>* Look_Ahead_Ptr;
	
	// Allocate the row pointers
	this->TheSparseMatrix_CreateMatrix = 
	  (void**)Mymalloc ((this->Num_equations) * sizeof(Node<T>*));
	assert_not_null(this->TheSparseMatrix_CreateMatrix);
	
	// initialize the pointers
	int i;
	for (i=0; i < this->Num_equations; i++)
	  {
		this->TheSparseMatrix_CreateMatrix[i] = NULL;
	  }
	
	Eqn_list_node<T>* temp = (Eqn_list_node<T>*)this->First_eqn_list_node;
	unsigned int eq_number = 0;
	unsigned int col;
	while (temp != NULL) 
	  {
		i = 0;
		if (temp->basis_pairs == NULL)
		  {
			return (1);
		  }
		while (temp->basis_pairs[i].coef != (T)0) 
		  {
			col = GetCol (temp->basis_pairs[i].left_basis,
						  temp->basis_pairs[i].right_basis);
			Tmp_S_Sum = (T)0;
			
			// since using a singly linked list we need ptr from the node before
			// this row,col value which is what Locate_Node will return. 
			// Locate_Node will try to return the one just before the column 
			// node that we want
			Prev_Ptr = Locate_Node<T> 
			  ((Node<T>**)this->TheSparseMatrix_CreateMatrix, eq_number, col);
			
			// if there is not one and the row is not empty this row,col value 
			// must be the first one in the row
			if ((Prev_Ptr == NULL) && 
				(! Row_empty<T> 
				 ((const Node<T>**)this->TheSparseMatrix_CreateMatrix, 
				  eq_number)))
			  {
				Look_Ahead_Ptr = 
				  ((Node<T>**)this->TheSparseMatrix_CreateMatrix)[eq_number];
			  }	
			else
			  {
				if (Row_empty<T> 
					((const Node<T>**)this->TheSparseMatrix_CreateMatrix, 
					 eq_number))
				  {
					// the row is empty
					Look_Ahead_Ptr = NULL;
				  }
				else
				  {
					// There is node after this one so we know that is that we 
					// want to work with
					if (Prev_Ptr->Next_Node != NULL)
					  {		
						Look_Ahead_Ptr = Prev_Ptr->Next_Node;
					  }
					else
					  {
						// there is not a node after this one so this is the one
						// that we want
						Look_Ahead_Ptr = Prev_Ptr;
					  }
				  }
			  }
			
			// Look_Ahead Ptr will only be null if there are NO nodes in the row
			if (Look_Ahead_Ptr != NULL)
			  {
				if (Look_Ahead_Ptr->column != col)
				  {	
					// Here we should add a node since there is no node in the 
					// row with the column value we are looking for
					Tmp_S_Sum = temp->basis_pairs[i].coef;
					if (Tmp_S_Sum != (T)0)
					  {
						this->Insert_Element<T> 
						  ((Node<T>**)this->TheSparseMatrix_CreateMatrix, 
						   eq_number, Tmp_S_Sum, col, Prev_Ptr);
						// if the flag is set increment # of elements for stat 
						// purpose
						if (this->gather_density_flag)
						  {
							this->num_elements++;
						  }
					  }
				  }
				else
				  {
					// There is a node here with the same column value we are
					// looking for so add the new value to this one
					Tmp_S_Sum = 
					  (Look_Ahead_Ptr->element) + (temp->basis_pairs[i].coef);
					if (Tmp_S_Sum == (T)0)
					  {
						// If the result is zero then we will want to delete 
						// this node
						this->Delete_Element<T> 
						  ((Node<T>**)this->TheSparseMatrix_CreateMatrix, 
						   eq_number, Prev_Ptr);
						// gather statistical information
						if (this->gather_density_flag)
						  {
							this->num_elements--;
						  }
					  }
					else
					  {
						// the result was nonzero and we just change the node 
						// element field to the result
						Change_Element<T> (Look_Ahead_Ptr, Tmp_S_Sum);
					  }
				  }
			  }
			// We know that there were no nonzero elements in the row so lets 
			// add one
			else
			  {
				this->Insert_Element<T> 
				  ((Node<T>**)this->TheSparseMatrix_CreateMatrix, eq_number,
				   temp->basis_pairs[i].coef, col, Prev_Ptr);

				// gather statistics
				if (this->gather_density_flag)
				  {
					this->num_elements++;
				  }
			  }	
			i++;
		  } // end while (temp->basis_pairs[i].coef != (T)0) 

		eq_number++;
		temp = temp->next;
	  }
	return (1);
  } // end of Albert::SparseFillTheMatrix()






  ////////////////////////////////////////////////////////////////////////////
  //                       SparseInterchange                                //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> 
  void SparseInterchange (const unsigned int Row1, const unsigned int Row2)
  {
	// switch the two row pointers
	Node<T>* Temp_Row_Ptr = ((Node<T>**)this->Matrix_Base_Ptr)[Row1];
	this->Matrix_Base_Ptr[Row1] = this->Matrix_Base_Ptr[Row2];
	this->Matrix_Base_Ptr[Row2] = Temp_Row_Ptr;
  }




  ////////////////////////////////////////////////////////////////////////////
  //                       SparseProcessDependentBasis                      //
  ////////////////////////////////////////////////////////////////////////////
  // Again this is virtually identical to the sister routine of 
  // ProcessDependentBasis() except we get the information from the sparse 
  // matrix structure. Also we do not use the dependent structure since the 
  // first element is each linked list representing a row is a nonzero element.
  template<typename T> void SparseProcessDependentBasis()
  {
	std::vector<std::pair<T, Basis> > tl;
	Node<T>* Tmp_Ptr;
	Node<T>* q;
	short int j, k;
	Basis b1, b2;
	unsigned int rowid = 0;
	while (rowid < this->MatrixRank)
	  {
		Tmp_Ptr = ((Node<T>**)this->TheSparseMatrix)[rowid];
		j = Tmp_Ptr->column;
		b1 = this->ColtoBP[j].left_basis;
		b2 = this->ColtoBP[j].right_basis;
		tl.clear();
		q = NULL;
		if (Tmp_Ptr->Next_Node != NULL)
		  {
			q = Tmp_Ptr->Next_Node;
		  }
		while (q)
		  {
			k = q->column;
			tl.push_back (std::make_pair 
						  (- Get_Matrix_Element<T> 
						   ((Node<T>**)this->TheSparseMatrix, rowid, k), 
						   this->BasisNames[k]));
			q = q->Next_Node;	
		  }
		this->EnterProduct<T> (b1, b2, tl);
		rowid++;
		Tmp_Ptr = ((Node<T>**)this->TheSparseMatrix)[rowid];
	  }
  } // end of Albert::SparseProcessDependentBasis()






  ////////////////////////////////////////////////////////////////////////////
  //                       SparseReduceMatrix                               //
  ////////////////////////////////////////////////////////////////////////////
  // called once, from SolveEquations()
  template<typename T> bool SparseReduceMatrix 
  (const int Num_rows_arg, const int Num_cols_arg, unsigned int* Rank)
	{
	  if ((Num_rows_arg == 0) || (Num_cols_arg == 0))
		{
		  return (1);
		}
	  assert_not_null(this->Matrix_Base_Ptr);
  
	  // Search for the rightmost nonzero element
	  // Dependent on the current stairrow
	  int j, col;
	  int nextstairrow = 0;
	  T x, Factor;
	  for (int i=0; i < Num_cols_arg; i++)
		{
		  for (j = nextstairrow; j < Num_rows_arg; j++)
			{
			  if (Get_Matrix_Element<T> 
				  ((Node<T>**)this->Matrix_Base_Ptr, j, i) != (T)0)
				{
				  break;
				}
			}
		  // When found interchange and then try to knockout any nonzero
		  // elements in the same column
		  if (j < Num_rows_arg)
			{
			  this->SparseInterchange<T> (nextstairrow, j);
			  
			  // code from SparseKnockOut()
			  col = i;
			  x = Get_Matrix_Element<T> 
				((Node<T>**)this->Matrix_Base_Ptr, nextstairrow, col);
			  if (x != (T)1)
				{
				  // if the rightmost element in the current row is not one 
				  // then multiply (code from SparseMultRow())
				  
				  // Get the first node in the row
				  Node<T>* PMR_Ptr = 
					((Node<T>**)this->Matrix_Base_Ptr)[nextstairrow];
				  
				  // Step thru row ... multiplying each element by the factor
				  Factor = S_inv(x);
				  while (PMR_Ptr != NULL)
					{
					  PMR_Ptr->element = (PMR_Ptr->element) * Factor;
					  PMR_Ptr = PMR_Ptr->Next_Node;
					}
				}
			  // try to knockout elements in column in the rows above
			  for (j=0; j < nextstairrow; j++)
				{
				  SparseAddRow<T> 
					(- Get_Matrix_Element<T>
					 ((Node<T>**)this->Matrix_Base_Ptr, j, col), 
					 nextstairrow, j);
				}
			  // try to knockout elements in column in the rows below
			  for (j = nextstairrow+1; j < Num_rows_arg; j++)
				{
				  SparseAddRow<T>
					(- Get_Matrix_Element<T> 
					 ((Node<T>**)this->Matrix_Base_Ptr, j, col), 
					 nextstairrow, j);
				}
			  /////////// end of code from SparseKnockut() ///////////////
				
			  nextstairrow++;
			}
		}
	  *Rank = nextstairrow;
	  return (1);
	} // end of Albert::SparseReduceMatrix()






  ////////////////////////////////////////////////////////////////////////////
  //                       SplitJthType                                     //
  ////////////////////////////////////////////////////////////////////////////
  // SplitJthType() and Gen() call each other recursively
  template<typename T> void SplitJthType (const int j)
  {
	if (this->status_Multpart != 1)
	  {
		return;
	  }
	if (j < this->Num_vars)
	  {
		this->Gen<T> (this->Var_types[j], this->Deg_var[j], j);
	  }
	else if (this->OKSetPartitions()) 
	  {
#if DEBUG_SET_PARTITIONS
		PrintSetPartitions();
#endif
		// Now create a substitution record for current set partition
		this->Substitution.resize ((this->Num_vars) * (this->Max_deg_var));

		// Start of recursive call
		this->status_Multpart = this->DoCreateSubs<T> (0, 0);

		//free (this->Substitution);
		this->Substitution.clear();
	  }
  } // end of Albert::SplitJthType()






  ////////////////////////////////////////////////////////////////////////////
  //                       Sub                                              //
  ////////////////////////////////////////////////////////////////////////////
  template<typename T> bool Sub (typename Alg_element<T>::type& Ans, 
								 term_node* W)
  {
	assert_not_null(W);
	if ((W->left == NULL) && (W->right == NULL)) 
	  {
		const int var_number = this->GetVarNumber (W->letter) - 1;
		SetAE<T> 
		  (Ans, 
		   this->Substitution[var_number * (this->Max_deg_var) + (this->Permutation_list[var_number])[W->number - 1] - 1], 
		   (T)1);
	  }
	else 
	  {
		typename Alg_element<T>::type left, right;
		this->Sub<T> (left,  W->left); 		
		this->Sub<T> (right, W->right);
		
		// This is where we use the multiplication table
		this->MultAE<T> (left, right, Ans);
	  }
	return (1);
  } // end of Albert::Sub()





  ////////////////////////////////////////////////////////////////////////////
  //                       SubstituteWord                                   //
  ////////////////////////////////////////////////////////////////////////////
  //
  // THE HEART OF THE MATTER. WE HAVE REACHED THE CORE.
  // THE WHOLE IDEA OF DYNAMIC PROGRAMMING IS EMBEDDED IN THIS ROUTINE.
  // Form one term of the equation corresponding to the term in the
  // identity. 
  //
  template<typename T> bool SubstituteWord (struct term_node* W)
  {
	if (W == NULL)
	  {
		return (1);
	  }
	typename Alg_element<T>::type ae1, ae2;
	
	this->Sub<T> (ae1, W->left);  // We can expand left tree of W
	this->Sub<T> (ae2, W->right); // We can expand the right tree of W
	
	// We can't do any more expansion. i.e We can't multiply ae1 & ae2.
	// Because we are entering new basis elements of degree of W.
	// But now it is time for new basis pairs.
	// The equations are nothing but summation of basis pairs.
	T alpha, beta;
	Basis_pair_node<T>* temp_list = NULL;

	typename std::map<Basis, T>::const_iterator ae1i, ae2i;
	ae1i = ae1.begin();
	while (ae1i != ae1.end()) 
	  {
		if (ae1i->first != 0)
		  {
			alpha = ae1i->second;
			if (alpha != (T)0)
			  {
				ae2i = ae2.begin();
				while (ae2i != ae2.end())
				  {
					if (ae2i->first != 0) 
					  {
						beta = ae2i->second; 
						if (beta != (T)0)
						  {
							if (this->running_list == NULL)
							  {
								temp_list = GetNewBPNode<T>(); 
								assert_not_null (temp_list);
								this->running_list = temp_list;
							  }
							else 
							  {
								temp_list->next = GetNewBPNode<T>();
								temp_list = temp_list->next;
								assert_not_null (temp_list);
							  }
							(temp_list->bp).coef = alpha * beta;
							(temp_list->bp).left_basis  = ae1i->first;
							(temp_list->bp).right_basis = ae2i->first;
						  }
					  }
					ae2i++;
				  }
			  }
		  }
		ae1i++;
	  }
	return (1);
  } // end of Albert::SubstituteWord()


}; // end of class Albert


#endif // __ALBERT_H
// end of albert.hh
