/*
  file:         driver.cc
  purpose:      This is the Main module of Albert. It does initializations.
                Interprets and executes Commands. Gets Command line and executes
                the command.
  created:      Sekhar Muddana circa 1990
  modified:     pasha may 17-19,jun 16,18-19,22,26 2019
  modification: ver. 4.0M7: 
                - openbsd support
				- all char* converted to std::string
				- changed argument handling (-f instead of -a)
				- started to add support for number fields
				- clear_mult_table() -> destroy_mult_table()
*/

#include <gmpxx.h>
#include <stdio.h>   // required by GNU readline

// required by readline on openbsd
#if __OpenBSD__
#define HAVE_ISXDIGIT
#endif

#include <readline/readline.h>
#include <readline/history.h>

#include <fstream>
#include "albert.hh"
#include "GF2.hh"
#include "GFp.hh"
#include "lidia_alg_num.hh"
#include "Ty_routines.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

static size_t __getline (std::string&);
static bool Compatible (const polynomial*, const P_type);
static bool Substr (const std::string&, const std::string&);

bool problem_type_present = 0;



int main (int argc, char* argv[])
{
  // argument handling
  // accepted arguments:
  // albert             argc = 1
  // albert -h          argc = 2
  // albert -f <file>   argc = 3
  if ((argc == 2) && (strcmp (argv[1], "-h") == 0))
	{
	  std::cout 
		<< "Usage: " << argv[0] << " [-f <filename>|-h]" << std::endl 
		<< "If the optional -f option is specified, use configuration file <filename>" << std::endl
		<< "-h option displays this help message and quits." << std::endl;
	  return (-1);
	}
  if ((argc != 1) && ((argc != 3) || (strcmp (argv[1], "-f") != 0)))
	{
	  std::cout << "Wrong arguments; try " << argv[0] << " -h" << std::endl;
	  return (-1);
	}
  // end of argument handling

  std::cout << std::endl 
			<< "Albert version 4.0M7, 2019"
			<< std::endl << std::endl;

  Albert A;

  // albert was invoked either without arguments, or with "-f <filename>",
  // in which case the third argument should be the name of the configuration 
  // file
  if (argc > 1)
	{
	  A.read_config_file (argv[2], 1);
	}
  
  // initialize the problem type
  int i;
  P_type ptype;
  for (i=0; i<NUM_LETTERS; i++)
    {
      ptype.degrees[i] = 0;
    }
  ptype.tot_degree = 0;

  // support for save and view commands
  char tableFileName[100];
  char table;
  //extern Deg_to_basis_rec* Deg_to_basis_table;
  std::filebuf tableFilePtr;

  std::string Command, Operand, 
	temp_line, temp_line2, total_line, total_line1, str1, str2, str3;

  polynomial* Cur_poly;
  P_type Temp_type;   // structure for problem type
  char c;
  int Cur_id_no, temp_line_len, tl;
  bool first_generator, match, slashfound;

  rl_bind_key ('\t', rl_insert);  // disable default tab completion in readline

  bool quit = 0;
  while (! quit)
    {
      std::cout << std::endl;
	  
      ////////// get the command and the operand ///////////////////////
      //
	  // Also uses the definitions in the config file (if supplied)
	  //
	  total_line.resize (0);
      temp_line_len = __getline (temp_line);
	  temp_line2.resize (0);

      slashfound = 1;
      while (slashfound && (temp_line_len > 0)) 
		{
		  i = 0;
		  while (temp_line[i] == ' ')
			{ 
			  // read past leading white space
			  ++i;
			}
		  while ((temp_line[i] != '\\') && (temp_line[i] != '\0'))
			{
			  temp_line2.push_back (temp_line[i++]);
			}
		  if (temp_line[i] == '\\') 
			{
			  slashfound = 1;
			  temp_line_len = __getline (temp_line);
			}
		  else
			{
			  slashfound = 0;
			}
		  total_line += temp_line2;
		}
      
      if (temp_line_len == -1)
		{
		  // error reading stdin, abort
		  goto END;
		}
	  
	  tl = total_line.length();
	  total_line1.resize (0);
      i = 0;
      while (i < tl)
		{
		  if ((total_line[i] == ' ') || (total_line[i] == '\t'))
			{
			  break;
			}
		  total_line1.push_back (total_line[i]);
		  i++;
		}
	  i++;
	  Command = total_line1;

	  total_line1.resize (0);
      while (i < tl) 
		{
		  if ((total_line[i] != ' ') && (total_line[i] != '\t'))
			{
			  total_line1.push_back (total_line[i++]);
			}
		  else
			{
			  i++;
			}
		}
	  total_line = total_line1;
	  
      if (! A.Dalbertheadptr.empty())
		{
		  while (1) 
			{
			  // Scans the total_line and breaks it into 3 strings str1, str2 
			  // and str3. str1 is the prefix before $ in total_line. str2 is 
			  // the string between 2 $'s. str3 is the tail. Now str2 is 
			  // substituted with a string as per the definitions in .albert. 
			  // i.e if str2 matches with some left hand side of a definition in
			  // .albert, then it is replaced with corresponding right hand 
			  // side.
			  //tnodeptr = A.Dalbertheadptr;
			  std::list<dalbert_node>::iterator it = A.Dalbertheadptr.begin();

			  tl = total_line.length();
			  if (tl < 3)
				{
				  break;
				}
			  str1.resize (0);
			  i = 0;
			  while ((i<tl) && (total_line[i] != '$'))
				{
				  str1.push_back (total_line[i++]);
				}
			  if (total_line[i++] != '$') 
				{
				  break;
				}
			  str2.resize (0);
			  while ((i<tl) && (total_line[i] != '$'))
				{
				  str2.push_back (total_line[i++]);
				}
			  if (total_line[i++] != '$') 
				{
				  break;
				}
			  str3.resize (0);
			  while (i<tl)
				{
				  str3.push_back (total_line[i++]);
				}
			  
			  match = 0;
			  while ((it != A.Dalbertheadptr.end()) && (! match))
				{
				  if (it->lhs.compare (str2) == 0)
					{
					  match = 1;
					  break;
					}
				  ++it;
				}
			  if (! match) 
				{
				  break;
				}
			  total_line.resize (0);
			  total_line += str1;
			  total_line += it->rhs;
			  total_line += str3;
#if DB_READ_ALBERT
			  std::cout << "str1 = " << str1 << std::endl 
						<< "str2 = " << str2 << std::endl
						<< "str3 = " << str3 << std::endl
						<< "total line = " << total_line << std::endl;
#endif
    	    } // end of substitution loop
		}
	  
	  Operand = total_line;
#if DB_READ_ALBERT
      std::cout << "command = " << Command << std::endl
				<< "operand = " << Operand << std::endl;
#endif
      ////////// end of get the command and the operand ////////////////
      
	  
      std::cout << std::endl;
      switch (Command[0]) 
		{
		case 'i':
		  // "identity polynomial". Accept new identity.
		  if (! Substr (Command, "identity")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  Cur_id_no = A.add_identity (Operand.c_str());
		  if (Cur_id_no > 0)
			{
			  Print_poly (A.Cur_id, A.Cur_poly_len);
			  std::cout << std::endl << "Entered as identity " << Cur_id_no 
						<< "." << std::endl;
			}
		  break;
		  
		case 'r':
		  // "remove number|*". remove identity.
		  if (! Substr (Command, "remove")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (Operand[0] == '*') 
			{
			  Remove_all_ids (&A.Id_queue);
			  std::cout << "Removed all the identities." << std::endl;
			  A.destroy_mult_table();
			}
		  else 
			{
			  Cur_id_no = atoi (Operand.c_str());
			  if (Cur_id_no > 0) 
				{
				  if (Remove_id (Cur_id_no, &A.Id_queue)) 
					{
					  std::cout << "Identity " << Cur_id_no << " removed." 
								<< std::endl;
					  A.destroy_mult_table();
					}
				  else
					{
					  std::cout << "Identity " << Cur_id_no << " not present.";
					}
				}
			  else
				{
				  std::cout << "Number must be at least 1.";
				}
			}
		  break;
		  
		case 'g':
		  // "generators word". Accept problem type.
		  if (! Substr (Command, "generators")) 
			{
			  std::cout << "Illegal command." << std::endl;
			  break;
			}
		  if (! Parse_generator_word (Operand.c_str(), &Temp_type))
			{
			  std::cout << "Error in the generator list." << std::endl;
			}
		  else if (Temp_type.tot_degree < 2)
			{
			  std::cout << "Total degree must be at least 2." << std::endl;
			}
		  else 
			{
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  ptype.degrees[i] = 0;
				}
			  ptype.tot_degree = 0;
			  
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  ptype.degrees[i] = Temp_type.degrees[i];
				}
			  ptype.tot_degree = Temp_type.tot_degree;
			  if (problem_type_present) 
				{
				  std::cout << "Problem type changed." << std::endl;
				}
			  else
				{
				  std::cout << "Problem type stored." << std::endl;
				}
			  AssignBasisNumberstoLetters (ptype);

			  // struct P_type ptype is used to create Target type Type 
			  // problem_target_type
			  // If generators are 3a2b2c then Target type is 322.
			  unsigned int ttype_len = 1;
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  if (ptype.degrees[i] > 0)
					{
					  ttype_len++;
					}
				}
			  A.Target_type.resize (ttype_len);
			  unsigned int cur_tt_index = 0;
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  if (ptype.degrees[i] > 0)
					{
					  A.Target_type[cur_tt_index++] = ptype.degrees[i];
					}
				}
			  A.Target_type[cur_tt_index] = 0;
			  problem_type_present = 1;
			  A.destroy_mult_table();
			}
		  break;
		  
		case 'f':
		  // "field". Change the ground field
		  if (! Substr (Command, "field")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  //p = atoll (Operand.c_str());
		  A.set_field (Operand, 1);
		  break;
		  
		case 'b':
		  // "build". Build the multiplication table.
		  if (! Substr (Command, "build")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (! problem_type_present) 
			{
			  std::cout << "Problem Type not present." << std::endl 
						<< "Can't build Multiplication Table." << std::endl;
			  break;
			}
		  switch (A.field_elem_type)
			{
			case GF2_TYPE:
			  A.build<GF2_element> (1);
			  break;

			case GFP_TYPE:
			  A.build<GFp_element> (1);
			  break;

			case GMP_TYPE:
			  A.build<mpq_class> (1);
			  break;

			case LIDIA_TYPE:
			  A.build<alg_number> (1);
			  break;

			default:
			  std::cerr << "wrong field_elem_type=" << A.field_elem_type
						<< std::endl;
			  break;
			}
		  break;
		  
		case 'd':
		  // "display". Display the current problem configuration.
		  if (! Substr (Command, "display")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (A.Id_queue.first != NULL) 
			{
			  std::cout << "Defining identities are:" << std::endl;
			  Print_ids (A.Id_queue);
			}
		  else
			{
			  std::cout << "No defining identities." << std::endl;
			}
		  std::cout << "Ground field is ";
		  A.print_field (std::cout, 0);
		  std::cout << std::endl;
		  
		  // Print the Problem type in a good looking format!
		  if (problem_type_present) 
			{
			  first_generator = 1;
			  std::cout << "Problem type = [";
			  for (i=0; i<NUM_LETTERS; i++) 
				{
				  if (ptype.degrees[i] == 1) 
					{
					  c = 'a' + i;
					  if (! first_generator)
						{
						  std::cout << "," << c;
						}
					  else 
						{
						  first_generator = 0;
						  std::cout << c;
						}
					}
				  else if (ptype.degrees[i] > 1) 
					{
					  c = 'a' + i;
					  if (! first_generator)
						{
						  std::cout << "," << (unsigned short)ptype.degrees[i] << c;
						}
					  else 
						{
						  first_generator = 0;
						  std::cout << (unsigned short)ptype.degrees[i] << c;
						}
					}
				}
			  std::cout << "]; Total degree = " 
						<< (unsigned int)ptype.tot_degree << "." << std::endl;
			}
		  else
			{
			  std::cout << "Problem type not specified." << std::endl;
			}
		  
		  std::cout << "Multiplication table ";
		  if (! A.mtable_present)
			{
			  std::cout << "not ";
			}
		  std::cout << "present.";
		  break;
		  
		case 'p':
		  // "polynomial poly". Enquire if poly is an identity
		  if (! Substr (Command, "polynomial")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (! A.mtable_present) 
			{
			  std::cout << "Multiplication table not present.";
			  break;
			}
		  Cur_poly = Create_poly (Operand.c_str(), &A.Cur_poly_len);
		  if (Cur_poly == NULL) 
			{
			  std::cout << Operand << " is not a polynomial.";
			  break;
			}
		  if (! Homogeneous (Cur_poly)) 
			{
			  std::cout << "Given polynomial is not homogeneous.";
			  break;
			}
		  if (Cur_poly->degree < 2) 
			{
			  std::cout << "Total degree must be at least 2.";
			  break;
			}
		  if (! Compatible (Cur_poly, ptype)) 
			{
			  std::cout << "Polynomial type not a subtype of Target_type.";
			  break;
			}

		  switch (A.field_elem_type)
			{
			case GF2_TYPE:
			  if (A.IsIdentity<GF2_element> (Cur_poly))
				{
				  std::cout << "Polynomial is an identity.";
				}
			  else
				{
				  std::cout << "Polynomial is not an identity.";
				}
			  break;
			  
			case GFP_TYPE:
			  if (A.IsIdentity<GFp_element> (Cur_poly))
				{
				  std::cout << "Polynomial is an identity.";
				}
			  else
				{
				  std::cout << "Polynomial is not an identity.";
				}
			  break;
			  
			case GMP_TYPE:
			  if (A.IsIdentity<mpq_class> (Cur_poly))
				{
				  std::cout << "Polynomial is an identity.";
				}
			  else
				{
				  std::cout << "Polynomial is not an identity.";
				}

			  break;
			  
			default:
			  std::cerr << "impossible field element type " 
						<< A.field_elem_type 
						<< std::endl;
			  break;
			}
		
		  DestroyPoly (Cur_poly);
		  break;
		  
		case 'x':
		  // "xpand poly"
		  if (! Substr (Command, "xpand")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  Cur_poly = Create_poly (Operand.c_str(), &A.Cur_poly_len);
		  if (Cur_poly != NULL) 
			{
			  if (! Homogeneous (Cur_poly)) 
				{
				  std::cout << "Given polynomial is not homogeneous.";
				  DestroyPoly (Cur_poly);
				  break;
				}
			  else 
				{
				  Print_poly (Cur_poly, A.Cur_poly_len);
				  std::cout << std::endl;
				  DestroyPoly (Cur_poly);
				}
			}
		  else 
			{
			  std::cout << Operand << " is not a polynomial.";
			}
		  break;
		  
		  // "save" the multiplication table or the basis table to a file
		case 's':
		  if (! Substr (Command, "save"))
			{
			  std::cout << "Illegal command.";
			  break;
			}
	  
		  table = Operand[0];
		  switch (table) 
			{
			case 'b':
			  if (
				  A.mtable_present && (A.Basis_table.size() > 1) && 
				  (! A.Deg_to_basis_table.empty())
				  )
				{
				  std::cout << "File Name --> ";
				  fflush (stdout);
				  fgets (tableFileName, sizeof(tableFileName), stdin);
				  tableFileName[ strlen(tableFileName) -1 ] = 0;
				  std::cout << std::endl;
				  if (strlen (tableFileName))
					{
					  if (! tableFilePtr.open (tableFileName, std::ios::out))
						{
						  std::cout << "Unable to open file " 
									<< tableFileName << std::endl;
						  break;
						}
					}
				  else
					{
					  std::cout << "No file name was entered. Command aborted"
								<< std::endl;
					  break;
					}
				  std::ostream os (&tableFilePtr);
				  A.PrintBasisTable (os);
				  tableFilePtr.close();
				}
			  else
				{
				  std::cout << "Basis Table not present." << std::endl;
				}
			  break;
			  
			case 'm':
			  if (A.mtable_present)
				{
				  std::cout << "File Name --> ";
				  fflush (stdout);
				  fgets (tableFileName, sizeof(tableFileName), stdin);
				  tableFileName[strlen(tableFileName) - 1] = 0;
				  std::cout << std::endl;
				  if (strlen (tableFileName))
					{
					  if (! tableFilePtr.open (tableFileName, std::ios::out))
						{
						  std::cout << "Unable to open file "
									<< tableFileName << std::endl;
						  break;
						}
					}
				  else
					{
					  std::cout << "No file name was entered. Command aborted"
								<< std::endl;
					  break;
					}
				  std::ostream os (&tableFilePtr);
				  A.print_mtable (os, strstr (Operand.c_str(), "gap") != NULL);
				  tableFilePtr.close();
				}
			  else
				{
				  std::cout << "Multiplication Table not present" << std::endl;
				}
			  break;
			  
			default:
			  std::cout << "Invalid table type. Specify \"m\" or \"b\"" 
						<< std::endl;
			}
		  break;
		  
		  // "view" the multiplication table or the basis table on screen
		case 'v':
		  if (! Substr (Command, "view"))
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  
		  table = Operand[0];
		  switch (table)
			{
			case 'b':
			  if (
				  A.mtable_present && (A.Basis_table.size() > 1) && 
				  (! A.Deg_to_basis_table.empty())
				  )
				{
				  A.PrintBasisTable (std::cout);
				}
			  else
				{
				  std::cout << "Basis Table not present." << std::endl;
				}
			break;
	    
		  case 'm':
			if (A.mtable_present)
			  {
				A.print_mtable (std::cout, strstr (Operand.c_str(), "gap") != NULL);
			  }
			else
			  {
				std::cout << "Multiplication Table not present." << std::endl;
			  }
			break;
			
		  default:
			std::cout << "Invalid table type " << table << "." << std::endl
					  << "Table type must be \"m\" or \"b\"" << std::endl;
		  }
		  break;
	  
		case 't':
		  // "type nonassoc. word". Find the type of nonassoc. word
		  if (! Substr (Command, "type")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  Cur_poly = Create_poly (Operand.c_str(), &A.Cur_poly_len);
		  if (Cur_poly == NULL) 
			{
			  std::cout << Operand << " is not a nonassociative word.";
			  break;
			}
		  else if (Cur_poly->terms->next != NULL)
			{
			  std::cout << Operand << " is not a nonassociative word.";
			}
		  else 
			{
			  i = Assoc_number(Cur_poly->terms->term);
			  std::cout << "The association type of the word = " << i << ".";
			}
		  break;
	  
		case '#':
		  // comment; ignore it (for batch processing)
		  break;

		case 'q':
		  // "quit". Quit Albert.
		  if (! Substr (Command, "quit")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  quit = 1;
		  break;
		  
		case '\0':
		  break;
		  
		default :
		  std::cout << "Illegal command." << std::endl;
		  break;
        }
    } // end of the main loop
  
 END:
  return (0);
} // end of main()




// Check if the polynomial has the compatible problem type.
// i.e if generators are 3a2b2c then each term in a polynomial should have at 
// most 3a's,2b's and 2c's. Else we can't use the multiplication table to answer
// the query whether the poly is an identity or not. 
static bool Compatible (const polynomial* Poly, const P_type ptype)
{
  if (Poly == NULL)
	{
	  return (0);
	}
  for (unsigned int i=0; i<NUM_LETTERS; i++)
	{
	  if (Poly->deg_letter[i] > ptype.degrees[i])
		{
		  return (0);
		}
	}
  return (1);
}


// wrapper for readline functions;
// returns line length, or -1 if error reading stdin occured
static size_t __getline (std::string& s)
{
  char* line = readline ("-->");
  if (line == NULL)
    {
      // end of stdin; this should happen in a non-interacive session only
      return (-1);
    }
  if (*line)
    {
      add_history (line);
    }
  size_t line_len = strlen (line);
  s = line;
  free (line);
  return (line_len);
}



// Used to check if Str1 is a string which is a prefix of Str2
static bool Substr (const std::string& Str1, const std::string& Str2)
{
  size_t len_str1 = Str1.length();
  size_t len_str2 = Str2.length();
  size_t i = 0;
  while ((i<len_str1) && (i<len_str2)) 
	{
	  if (Str1[i] != Str2[i])
		{
		  return (0);
		}
	  else
		{
		  i++;
		}
    }
  return (i >= len_str1);
}


// end of driver.cc
